"""Test achievements."""
from mpf.tests.MpfFakeGameTestCase import MpfFakeGameTestCase


class TestAchievement(MpfFakeGameTestCase):
    def get_config_file(self):
        return 'config.yaml'

    def get_machine_path(self):
        return 'tests/machine_files/achievement/'

    def test_enable_when_all_done(self):
        self.start_game()
        self.mock_event("achievement_mode1_a1_state_selected")
        self.mock_event("achievement_mode1_a2_state_selected")
        self.mock_event("enable_all")
        self.advance_time_and_run()
        self.post_event("start_mode1")
        self.advance_time_and_run()
        self.assertEventCalled("enable_all", times=1)
        self.post_event("stop_mode1")

        self.post_event("start_all")
        self.post_event("complete_all")
        self.assertEqual("completed", self.machine.achievements['mode1_a1'].state)
        self.assertEqual("completed", self.machine.achievements['mode1_a2'].state)

        self.assertEventCalled("enable_all", times=1)
        self.post_event("start_mode1")
        self.advance_time_and_run()
        self.assertEventCalled("enable_all", times=2)

    def test_two_players_restart_with_keep(self):
        self.mock_event("achievement_achievement1_state_disabled")
        self.mock_event("achievement_achievement1_state_enabled")
        self.mock_event("achievement_achievement1_state_started")
        self.mock_event("achievement_achievement1_state_stopped")
        self.mock_event("achievement_achievement1_state_completed")

        achievement = self.machine.achievements['achievement1']

        self.assertEqual(None, achievement._show)

        self.start_two_player_game()
        self.assertModeRunning('base')
        self.advance_time_and_run(10)

        _, sub1 = self.machine.placeholder_manager.build_raw_template(
            "device.achievements.achievement1.state").evaluate_and_subscribe([])
        _, sub2 = self.machine.placeholder_manager.build_raw_template(
            "device.achievements.achievement2.state").evaluate_and_subscribe([])

        # start disabled
        self.assertEqual("achievement1_disabled", achievement._show.name)
        self.assertEqual("disabled", achievement.state)
        self.assertEqual(1, self._events['achievement_achievement1_state_disabled'])
        self.assertEqual(0, self._events['achievement_achievement1_state_enabled'])
        self.assertEqual(0, self._events['achievement_achievement1_state_started'])
        self.assertEqual(0, self._events['achievement_achievement1_state_stopped'])
        self.assertEqual(0, self._events['achievement_achievement1_state_completed'])
        self.assertFalse(self._last_event_kwargs['achievement_achievement1_state_disabled']['restore'])

        # should not start
        self.post_event("achievement1_start")
        self.assertEqual("disabled", achievement.state)
        self.assertPlaceholderEvaluates("disabled", "device.achievements.achievement1.state")
        self.assertPlaceholderEvaluates("disabled", "device.achievements.achievement2.state")
        self.assertFalse(sub1.done())
        self.assertFalse(sub2.done())

        # enable
        self.post_event("achievement1_enable")
        self.assertEqual("enabled", achievement.state)
        self.assertEqual(1, self._events['achievement_achievement1_state_disabled'])
        self.assertEqual(1, self._events['achievement_achievement1_state_enabled'])
        self.assertEqual(0, self._events['achievement_achievement1_state_started'])
        self.assertEqual(0, self._events['achievement_achievement1_state_completed'])
        self.assertPlaceholderEvaluates("enabled", "device.achievements.achievement1.state")
        self.assertEqual("achievement1_enabled", achievement._show.name)
        self.assertFalse(self._last_event_kwargs['achievement_achievement1_state_enabled']['restore'])
        self.assertTrue(sub1.done())
        self.assertFalse(sub2.done())

        self.post_event("achievement1_start")
        self.assertEqual("started", achievement.state)
        self.assertEqual(1, self._events['achievement_achievement1_state_disabled'])
        self.assertEqual(1, self._events['achievement_achievement1_state_enabled'])
        self.assertEqual(1, self._events['achievement_achievement1_state_started'])
        self.assertEqual(0, self._events['achievement_achievement1_state_completed'])
        self.assertPlaceholderEvaluates("started", "device.achievements.achievement1.state")
        self.assertEqual("achievement1_started", achievement._show.name)
        self.assertFalse(self._last_event_kwargs['achievement_achievement1_state_started']['restore'])

        _, sub1 = self.machine.placeholder_manager.build_raw_template(
            "device.achievements.achievement1.state").evaluate_and_subscribe([])

        self.drain_all_balls()

        self.assertPlayerNumber(2)
        self.assertBallNumber(1)
        self.assertEqual("achievement1_disabled", achievement._show.name)
        self.assertEqual("disabled", achievement.state)
        self.assertPlaceholderEvaluates("disabled", "device.achievements.achievement1.state")
        self.assertFalse(self._last_event_kwargs['achievement_achievement1_state_disabled']['restore'])
        self.assertTrue(sub1.done())

        self.post_event("achievement1_enable")
        self.assertEqual("achievement1_enabled", achievement._show.name)
        self.assertEqual("enabled", achievement.state)

        _, sub1 = self.machine.placeholder_manager.build_raw_template(
            "device.achievements.achievement1.state").evaluate_and_subscribe([])

        self.drain_all_balls()
        self.assertPlayerNumber(1)
        self.assertBallNumber(2)
        self.assertEqual("achievement1_started", achievement._show.name)
        self.assertEqual("started", achievement.state)
        self.assertTrue(self._last_event_kwargs['achievement_achievement1_state_started']['restore'])
        self.assertTrue(sub1.done())

        self.post_event("achievement1_complete")
        self.assertEqual("achievement1_completed", achievement._show.name)
        self.assertEqual("completed", achievement.state)
        self.assertFalse(self._last_event_kwargs['achievement_achievement1_state_completed']['restore'])

        self.post_event("achievement1_enable")
        self.post_event("achievement1_disable")
        self.post_event("achievement1_stop")
        self.assertEqual("completed", achievement.state)

        self.drain_all_balls()
        self.assertPlayerNumber(2)
        self.assertBallNumber(2)

        self.assertEqual("achievement1_enabled", achievement._show.name)
        self.assertEqual("enabled", achievement.state)
        self.assertTrue(self._last_event_kwargs['achievement_achievement1_state_enabled']['restore'])

        self.post_event("achievement1_disable")
        self.assertEqual("disabled", achievement.state)

        self.post_event("achievement1_enable")
        self.assertEqual("enabled", achievement.state)

        self.post_event("achievement1_start")
        self.assertEqual("started", achievement.state)

        self.post_event("achievement1_stop")
        self.assertEqual("stopped", achievement.state)
        self.assertEqual("achievement1_stopped", achievement._show.name)

        self.drain_all_balls()
        self.assertPlayerNumber(1)
        self.assertBallNumber(3)

        self.assertEqual("achievement1_completed", achievement._show.name)
        self.assertEqual("completed", achievement.state)
        self.assertTrue(self._last_event_kwargs['achievement_achievement1_state_completed']['restore'])

        self.post_event("achievement1_reset")
        self.assertEqual("achievement1_disabled", achievement._show.name)
        self.assertEqual("disabled", achievement.state)
        self.assertFalse(self._last_event_kwargs['achievement_achievement1_state_disabled']['restore'])

        self.drain_all_balls()
        self.assertPlayerNumber(2)
        self.assertBallNumber(3)

        self.assertEqual("stopped", achievement.state)
        self.assertEqual("achievement1_stopped", achievement._show.name)
        self.assertTrue(self._last_event_kwargs['achievement_achievement1_state_stopped']['restore'])

        self.post_event("achievement1_start")
        self.assertEqual("started", achievement.state)

        self.drain_all_balls()
        self.assertGameIsNotRunning()

        self.assertEqual(None, achievement._show)

    def test_one_player_no_restart(self):
        self.mock_event("test_event")
        self.mock_event("test_event2")
        achievement = self.machine.achievements['achievement2']
        self.assertEqual(None, achievement._show)
        self.assertLightColor('led1', 'off')

        self.start_game()

        # enable
        self.post_event('achievement2_enable', 1)

        self.assertEqual("achievement2_enabled", achievement._show.name)
        self.assertEqual("enabled", achievement.state)
        self.assertLightColor('led1', 'yellow')

        self.drain_all_balls()
        self.assertPlayerNumber(1)
        self.assertBallNumber(2)

        self.assertEqual(None, achievement._show)
        self.assertEqual("disabled", achievement.state)
        self.assertLightColor('led1', 'off')

        self.post_event("achievement2_enable", 2)
        self.assertEqual("enabled", achievement.state)
        self.assertLightColor('led1', 'yellow')

        self.assertEqual(0, self._events['test_event'])
        self.assertEqual(0, self._events['test_event2'])

        self.post_event("achievement2_start", 2)
        self.assertEqual("achievement2_started", achievement._show.name)
        self.assertEqual("started", achievement.state)
        self.assertLightColor('led1', 'green')

        self.assertEqual(1, self._events['test_event'])
        self.assertEqual(1, self._events['test_event2'])
        self.drain_all_balls()
        self.assertPlayerNumber(1)
        self.assertBallNumber(3)

        self.assertEqual("stopped", achievement.state)
        self.assertEqual(None, achievement._show)
        self.assertLightColor('led1', 'off')  # no show when stopped

        # restart after stop is False
        self.post_event("achievement2_start", 2)
        self.assertEqual("stopped", achievement.state)
        self.assertLightColor('led1', 'off')

    def test_control_events_in_mode(self):
        baseMode = self.machine.modes["base"]
        ach = self.machine.achievements['achievement7']
        group = self.machine.achievement_groups['group1']

        # The custom behavior is unique to mode devices _without_ enable_events
        self.assertFalse(ach.config['enable_events'])
        self.assertFalse(group.config['enable_events'])
        # Assert that the achievement does not add mode event handlers
        self.assertEqual(baseMode.event_handlers, set())
        ach.add_control_events_in_mode(mode=baseMode)
        self.assertEqual(baseMode.event_handlers, set())
        # Assert that achievement groups do add event handlers
        group.add_control_events_in_mode(mode=baseMode)
        self.assertEqual(len(baseMode.event_handlers), 1)

    def test_group_select(self):

        a4 = self.machine.achievements['achievement4']
        a5 = self.machine.achievements['achievement5']
        a6 = self.machine.achievements['achievement6']
        g2 = self.machine.achievement_groups["group2"]

        self.start_game()

        self.assertFalse(g2.enabled)
        self.assertEqual("disabled", a4.state)
        self.assertEqual("disabled", a5.state)
        self.assertEqual("disabled", a6.state)
        self.assertLightColor('led4', 'off')
        self.assertLightColor('led5', 'off')
        self.assertLightColor('led6', 'off')

        # achievements are disabled, should select none

        self.post_event('group2_random', 2)

        enabled_achievements = [x for x in (a4, a5, a6) if x.state == 'enabled']
        self.assertEqual(len(enabled_achievements), 0)

        # enable the achievements and try again, but group is disabled

        a4.enable()
        a5.enable()
        a6.enable()

        self.assertEqual("enabled", a4.state)
        self.assertEqual("enabled", a5.state)
        self.assertEqual("enabled", a6.state)
        self.advance_time_and_run(1)
        self.assertLightColor('led4', 'yellow')
        self.assertLightColor('led5', 'yellow')
        self.assertLightColor('led6', 'yellow')

        self.post_event('group2_random', 2)

        selected_achievements = [x for x in (a4, a5, a6) if x.state == 'selected']
        self.assertEqual(len(selected_achievements), 0)

        # now enable the group and it should work

        self.post_event('group2_enable', 2)
        self.assertTrue(g2._enabled)
        self.assertLightColors('led2', ['red', 'blue'], 1, .09)  # group led

        self.post_event('group2_random', 2)
        selected_achievements = [x for x in (a4, a5, a6) if x.selected]
        self.assertEqual(len(selected_achievements), 1)

        # group enabled, but individual members disabled, should not work

        a4.disable()
        a5.disable()
        a6.disable()
        self.advance_time_and_run(1)

        self.assertEqual("disabled", a4.state)
        self.assertEqual("disabled", a5.state)
        self.assertEqual("disabled", a6.state)
        self.assertLightColor('led4', 'off')
        self.assertLightColor('led5', 'off')
        self.assertLightColor('led6', 'off')

        self.post_event('group2_random')
        selected_achievements = [x for x in (a4, a5, a6) if x.state == 'selected']
        self.assertEqual(len(selected_achievements), 0)

    def test_rotation_when_all_complete(self):
        a14 = self.machine.achievements['achievement14']
        a15 = self.machine.achievements['achievement15']
        a16 = self.machine.achievements['achievement16']
        g4 = self.machine.achievement_groups["group4"]
        g4.enable()

        self.start_game()
        # complete all achievements
        a14.enable()
        a15.enable()
        a16.enable()
        a14.start()
        a15.start()
        a16.start()
        a14.complete()
        a14.complete()
        a15.complete()
        a16.complete()
        self.advance_time_and_run(.1)

        self.assertFalse(g4.enabled)
        self.advance_time_and_run()
        self.assertEqual("completed", a14.state)
        self.assertEqual("completed", a15.state)
        self.assertEqual("completed", a16.state)
        # rotate should not crash
        self.post_event('group4_rotate_right')

    def test_events_for_achievements(self):
        a17 = self.machine.achievements['achievement17']
        self.start_game()

        self.assertEqual("enabled", a17.state)

        # do not post enable event again
        self.mock_event("achievement_achievement17_state_enabled")
        a17.enable()
        self.machine_run()
        self.assertEventNotCalled("achievement_achievement17_state_enabled")

        # test disable
        self.mock_event("achievement_achievement17_state_disabled")
        a17.disable()
        self.machine_run()
        self.assertEventCalled("achievement_achievement17_state_disabled")

        # but only once
        self.mock_event("achievement_achievement17_state_disabled")
        a17.disable()
        self.machine_run()
        self.assertEventNotCalled("achievement_achievement17_state_disabled")

        # test enable
        self.mock_event("achievement_achievement17_state_enabled")
        a17.enable()
        self.machine_run()
        self.assertEventCalled("achievement_achievement17_state_enabled")

        _, sub = self.machine.placeholder_manager.build_raw_template(
            "device.achievements.achievement17.selected").evaluate_and_subscribe([])
        _, sub2 = self.machine.placeholder_manager.build_raw_template(
            "device.achievements.achievement16.selected").evaluate_and_subscribe([])

        # but only once
        self.mock_event("achievement_achievement17_state_enabled")
        a17.enable()
        self.machine_run()
        self.assertEventNotCalled("achievement_achievement17_state_enabled")
        self.assertPlaceholderEvaluates("enabled", "device.achievements.achievement17.state")
        self.assertPlaceholderEvaluates(False, "device.achievements.achievement17.selected")
        self.assertFalse(sub.done())
        self.assertFalse(sub2.done())

        # test select
        self.mock_event("achievement_achievement17_state_selected")
        a17.select()
        self.machine_run()
        self.assertEventCalled("achievement_achievement17_state_selected")
        self.assertPlaceholderEvaluates("enabled", "device.achievements.achievement17.state")
        self.assertPlaceholderEvaluates(True, "device.achievements.achievement17.selected")
        self.assertTrue(sub.done())
        self.assertFalse(sub2.done())

        # but only once
        self.mock_event("achievement_achievement17_state_selected")
        a17.select()
        self.machine_run()
        self.assertEventNotCalled("achievement_achievement17_state_selected")

        # test start
        self.mock_event("achievement_achievement17_state_started")
        a17.start()
        self.machine_run()
        self.assertEventCalled("achievement_achievement17_state_started")

        # but only once
        self.mock_event("achievement_achievement17_state_started")
        a17.start()
        self.machine_run()
        self.assertEventNotCalled("achievement_achievement17_state_started")

        # test complete
        self.mock_event("achievement_achievement17_state_completed")
        a17.complete()
        self.machine_run()
        self.assertEventCalled("achievement_achievement17_state_completed")

        # but only once
        self.mock_event("achievement_achievement17_state_completed")
        a17.complete()
        self.machine_run()
        self.assertEventNotCalled("achievement_achievement17_state_completed")

    def test_rotation(self):

        a4 = self.machine.achievements['achievement4']
        a5 = self.machine.achievements['achievement5']
        a6 = self.machine.achievements['achievement6']
        g2 = self.machine.achievement_groups['group2']

        self.start_game()

        a4.enable()
        a4.select()
        a5.enable()
        a6.enable()

        self.advance_time_and_run(1)
        self.assertEqual("enabled", a4.state)
        self.assertEqual("enabled", a5.state)
        self.assertEqual("enabled", a6.state)
        self.assertTrue(a4.selected)
        self.assertFalse(a5.selected)
        self.assertFalse(a6.selected)
        self.assertLightColor('led4', 'orange')
        self.assertLightColor('led5', 'yellow')
        self.assertLightColor('led6', 'yellow')

        # don't rotate if group is not enabled
        self.assertEqual(g2.enabled, False)
        self.post_event('group2_rotate_right', 1)
        self.assertEqual("enabled", a4.state)
        self.assertEqual("enabled", a5.state)
        self.assertEqual("enabled", a6.state)
        self.assertTrue(a4.selected)
        self.assertFalse(a5.selected)
        self.assertFalse(a6.selected)
        self.assertLightColor('led4', 'orange')
        self.assertLightColor('led5', 'yellow')
        self.assertLightColor('led6', 'yellow')

        # enable the group and test
        self.post_event('group2_enable', 1)
        self.assertEqual(g2.enabled, True)
        self.post_event('group2_rotate_right', 1)
        self.assertEqual("enabled", a4.state)
        self.assertEqual("enabled", a5.state)
        self.assertEqual("enabled", a6.state)
        self.assertFalse(a4.selected)
        self.assertTrue(a5.selected)
        self.assertFalse(a6.selected)
        self.assertLightColor('led4', 'yellow')
        self.assertLightColor('led5', 'orange')
        self.assertLightColor('led6', 'yellow')

        # rotate 2 more times to make sure
        self.post_event('group2_rotate_right', 1)
        self.assertEqual("enabled", a4.state)
        self.assertEqual("enabled", a5.state)
        self.assertEqual("enabled", a6.state)
        self.assertFalse(a4.selected)
        self.assertFalse(a5.selected)
        self.assertTrue(a6.selected)
        self.assertLightColor('led4', 'yellow')
        self.assertLightColor('led5', 'yellow')
        self.assertLightColor('led6', 'orange')

        self.post_event('group2_rotate_right', 1)
        self.assertEqual("enabled", a4.state)
        self.assertEqual("enabled", a5.state)
        self.assertEqual("enabled", a6.state)
        self.assertTrue(a4.selected)
        self.assertFalse(a5.selected)
        self.assertFalse(a6.selected)
        self.assertLightColor('led4', 'orange')
        self.assertLightColor('led5', 'yellow')
        self.assertLightColor('led6', 'yellow')

        # test rotate left
        self.post_event('group2_rotate_left', 1)
        self.assertEqual("enabled", a4.state)
        self.assertEqual("enabled", a5.state)
        self.assertEqual("enabled", a6.state)
        self.assertFalse(a4.selected)
        self.assertFalse(a5.selected)
        self.assertTrue(a6.selected)
        self.assertLightColor('led4', 'yellow')
        self.assertLightColor('led5', 'yellow')
        self.assertLightColor('led6', 'orange')

        # don't rotate between disabled ones
        self.post_event('achievement4_disable', 1)
        self.assertEqual("disabled", a4.state)
        self.assertEqual("enabled", a5.state)
        self.assertEqual("enabled", a6.state)
        self.assertFalse(a4.selected)
        self.assertFalse(a5.selected)
        self.assertTrue(a6.selected)
        self.assertLightColor('led4', 'off')
        self.assertLightColor('led5', 'yellow')
        self.assertLightColor('led6', 'orange')

        self.post_event('group2_rotate_right', 1)
        self.assertEqual("disabled", a4.state)
        self.assertEqual("enabled", a5.state)
        self.assertEqual("enabled", a6.state)
        self.assertFalse(a4.selected)
        self.assertTrue(a5.selected)
        self.assertFalse(a6.selected)
        self.assertLightColor('led4', 'off')
        self.assertLightColor('led5', 'orange')
        self.assertLightColor('led6', 'yellow')

        # complete the event, auto_select is not enabled

        self.post_event('achievement5_start', 1)
        self.post_event('achievement5_complete', 1)

        self.assertEqual("disabled", a4.state)
        self.assertEqual("completed", a5.state)
        self.assertEqual("enabled", a6.state)
        self.assertFalse(a4.selected)
        self.assertFalse(a5.selected)
        self.assertFalse(a6.selected)
        self.assertLightColor('led4', 'off')
        self.assertLightColor('led5', 'blue')
        self.assertLightColor('led6', 'yellow')

        # rotate should not touch the complete or disabled ones

        self.post_event('group2_rotate_right', 1)
        self.assertEqual("disabled", a4.state)
        self.assertEqual("completed", a5.state)
        self.assertEqual("enabled", a6.state)
        self.assertLightColor('led4', 'off')
        self.assertLightColor('led5', 'blue')
        self.assertLightColor('led6', 'yellow')
        self.assertFalse(a4.selected)
        self.assertFalse(a5.selected)
        self.assertFalse(a6.selected)

        # enable a4, select one, rotate shouldn't touch the disabled one
        self.post_event('achievement4_enable', 1)
        self.post_event('group2_random', 1)
        self.post_event('group2_rotate_right', 1)

        self.assertEqual("completed", a5.state)
        self.assertLightColor('led5', 'blue')

        if a4.selected:
            old_selected = a4
            old_enabled = a6
            self.assertLightColor('led4', 'orange')
            self.assertLightColor('led6', 'yellow')
            self.assertEqual("enabled", a6.state)
        else:
            old_selected = a6
            old_enabled = a4
            self.assertLightColor('led4', 'yellow')
            self.assertLightColor('led6', 'orange')
            self.assertEqual("enabled", a4.state)

        self.post_event('group2_rotate_right', 1)
        self.assertLightColor(old_selected.config['show_tokens']['led'], 'yellow')
        self.assertLightColor(old_enabled.config['show_tokens']['led'], 'orange')
        self.assertEqual(old_selected.state, 'enabled')
        self.assertEqual(old_enabled.state, 'enabled')
        self.assertFalse(old_selected.selected)
        self.assertTrue(old_enabled.selected)

        self.post_event('group2_rotate_right', 1)
        self.assertLightColor(old_selected.config['show_tokens']['led'], 'orange')
        self.assertLightColor(old_enabled.config['show_tokens']['led'], 'yellow')
        self.assertEqual(old_selected.state, 'enabled')
        self.assertEqual(old_enabled.state, 'enabled')
        self.assertTrue(old_selected.selected)
        self.assertFalse(old_enabled.selected)

    def test_group_completion_via_methods(self):
        a4 = self.machine.achievements['achievement4']
        a5 = self.machine.achievements['achievement5']
        a6 = self.machine.achievements['achievement6']
        g2 = self.machine.achievement_groups['group2']

        self.mock_event('group2_complete')
        self.mock_event('group2_no_more')

        self.start_game()

        _, sub = self.machine.placeholder_manager.build_raw_template(
            "device.achievement_groups.group2.enabled").evaluate_and_subscribe([])

        self.assertPlaceholderEvaluates(False, "device.achievement_groups.group2.enabled")

        # test via methods
        a4.enable()
        a5.enable()
        a6.enable()
        g2.enable()
        self.machine_run()

        self.assertPlaceholderEvaluates(True, "device.achievement_groups.group2.enabled")
        self.assertTrue(sub.done())

        self.assertEqual("enabled", a4.state)
        self.assertEqual("enabled", a5.state)
        self.assertEqual("enabled", a6.state)

        a4.select()
        self.assertTrue(a4.selected)
        self.assertEqual("enabled", a4.state)
        a4.start()
        self.assertEqual("started", a4.state)
        a4.stop()
        self.assertEqual("stopped", a4.state)
        a4.start()
        self.assertEqual("started", a4.state)
        a4.complete()
        self.assertEqual("completed", a4.state)

        a5.start()
        a5.complete()
        self.advance_time_and_run()
        self.assertEventNotCalled('group2_no_more')

        a6.start()
        self.advance_time_and_run()
        self.assertEventCalled('group2_no_more')
        self.assertEventNotCalled('group2_complete')

        a6.complete()
        self.advance_time_and_run()
        self.assertEventCalled('group2_complete')

        # re-enable & select random, should not select any

        g2.enable()
        g2.select_random_achievement()
        self.assertEqual("completed", a4.state)
        self.assertEqual("completed", a5.state)
        self.assertEqual("completed", a6.state)

    def test_group_auto_select_and_group_auto_enable(self):
        g1 = self.machine.achievement_groups['group1']

        self.start_game()

        self.assertTrue(self.machine.achievement_groups["group1"].enabled)
        selected = g1._selected_member
        self.assertTrue(selected)

        # group should auto disable
        selected.start()
        self.advance_time_and_run(.1)
        self.assertFalse(g1.enabled)

        # group should auto enable
        selected.stop()
        self.advance_time_and_run(.1)
        self.assertTrue(g1.enabled)

        _, sub = self.machine.placeholder_manager.build_raw_template(
            "device.achievement_groups.group1.selected_member").evaluate_and_subscribe([])

        selected.start()
        self.advance_time_and_run(.1)
        self.assertFalse(g1.enabled)
        selected.complete()
        self.advance_time_and_run(.1)

        # group should auto enable and select another
        self.assertTrue(g1.enabled)
        selected2 = g1._selected_member
        self.assertIsNot(selected, selected2)

        self.assertTrue(sub.done())

        selected2.start()
        self.advance_time_and_run(.1)
        selected2.complete()
        self.advance_time_and_run(.1)

        # group should auto enable and select another
        self.assertTrue(g1.enabled)
        selected3 = g1._selected_member
        self.assertIsNot(selected2, selected3)

        selected3.start()
        self.advance_time_and_run(.1)
        selected3.complete()
        self.advance_time_and_run(.1)

        # should not re-enable since all members are complete
        self.assertFalse(g1.enabled)

    def test_auto_enable_with_no_enable_events(self):
        self.start_game()

        self.assertTrue(self.machine.achievement_groups["group1"].enabled)
        self.assertFalse(self.machine.achievement_groups["group2"].enabled)

    def test_auto_select_with_no_enable_events(self):
        # a10 and 11 do not have enable events, so they should be enabled on
        # start. a12 and 13 have enable events, so they should not be enabled.
        # initial selection should pick either 10 or 11

        a10 = self.machine.achievements['achievement10']
        a11 = self.machine.achievements['achievement11']
        a12 = self.machine.achievements['achievement12']
        a13 = self.machine.achievements['achievement13']

        self.start_game()
        self.advance_time_and_run(.1)

        self.assertEqual(a10.state, 'enabled')
        self.assertEqual(a11.state, 'enabled')
        self.assertEqual(a12.state, 'disabled')
        self.assertEqual(a13.state, 'disabled')
        if a10.selected == a11.selected:
            raise AssertionError("Neither a10 nor a11 is selected")

    def _assert_selected(self, expected):
        self.assertEqual(
            expected,
            self.machine.achievements["spinTasticAward"].selected +
            self.machine.achievements["tagTeamAward"].selected +
            self.machine.achievements["doubleChanceAward"].selected +
            self.machine.achievements["extraBallAward"].selected +
            self.machine.achievements["prodigiousPopsAward"].selected,
            "Expected exactly {} achievement(s) to be selected. Found: {} {} {} {} {}".format(
                expected,
                self.machine.achievements["spinTasticAward"].selected,
                self.machine.achievements["tagTeamAward"].selected,
                self.machine.achievements["doubleChanceAward"].selected,
                self.machine.achievements["extraBallAward"].selected,
                self.machine.achievements["prodigiousPopsAward"].selected
            )
        )

    def test_auto_select_with_allow_selection_change_while_disabled(self):
        self.start_game()
        self.start_mode("auto_select")
        self._assert_selected(0)

        self.post_event("enable_group")
        self._assert_selected(1)
        self.post_event("start_event")
        self._assert_selected(1)
        self.post_event("disable_bonus")
        self._assert_selected(1)
        self.post_event("sw_pops")
        self._assert_selected(1)
        self.post_event("mode_spinTasticAward_stopped")
        self._assert_selected(1)
        self.post_event("mode_tagTeamAward_stopped")
        self._assert_selected(1)
        self.post_event("mode_doubleChanceAward_stopped")
        self._assert_selected(1)
        self.post_event("extraBallAwardIntro_complete")
        self._assert_selected(1)
        self.post_event("mode_prodigiousPopsAward_stopped")
        self.advance_time_and_run(.1)
        self._assert_selected(1)



"""FAST hardware platform.

Contains the hardware interface and drivers for the FAST Pinball platform
hardware, including the FAST Core and WPC controllers as well as FAST I/O
boards.
"""
import asyncio
import os
from copy import deepcopy
from distutils.version import StrictVersion
from typing import Dict, Set, Optional
import serial.tools.list_ports

from mpf.platforms.fast.fast_io_board import FastIoBoard
from mpf.platforms.fast.fast_servo import FastServo
from mpf.platforms.fast import fast_defines
from mpf.platforms.fast.fast_dmd import FASTDMD
from mpf.platforms.fast.fast_driver import FASTDriver
from mpf.platforms.fast.fast_gi import FASTGIString
from mpf.platforms.fast.fast_led import FASTDirectLED, FASTDirectLEDChannel
from mpf.platforms.fast.fast_light import FASTMatrixLight
from mpf.platforms.fast.fast_serial_communicator import FastSerialCommunicator
from mpf.platforms.fast.fast_switch import FASTSwitch

from mpf.core.platform import ServoPlatform, DmdPlatform, SwitchPlatform, DriverPlatform, LightsPlatform, \
    DriverSettings, SwitchSettings, DriverConfig, SwitchConfig, RepulseSettings
from mpf.core.utility_functions import Util


# pylint: disable-msg=too-many-instance-attributes
from mpf.platforms.interfaces.light_platform_interface import LightPlatformInterface


class FastHardwarePlatform(ServoPlatform, LightsPlatform, DmdPlatform,
                           SwitchPlatform, DriverPlatform):

    """Platform class for the FAST hardware controller."""

    __slots__ = ["dmd_connection", "net_connection", "rgb_connection", "serial_connections", "fast_leds",
                 "flag_led_tick_registered", "config", "machine_type", "hw_switch_data", "io_boards", "fast_commands",
                 "_watchdog_task", "_led_task"]

    def __init__(self, machine):
        """Initialise fast hardware platform.

        Args:
        ----
            machine: The main ``MachineController`` instance.
        """
        super().__init__(machine)

        self.config = self.machine.config_validator.validate_config("fast", self.machine.config['fast'])
        self._configure_device_logging_and_debug("FAST", self.config)

        if self.config["driverboards"]:
            self.machine_type = self.config["driverboards"]
        else:
            self.machine_type = self.machine.config['hardware']['driverboards'].lower()

        if self.machine_type == 'wpc':
            self.debug_log("Configuring the FAST Controller for WPC driver "
                           "board")
        else:
            self.debug_log("Configuring FAST Controller for FAST IO boards.")

        self.features['tickless'] = True

        self.dmd_connection = None
        self.net_connection = None
        self.rgb_connection = None
        self._watchdog_task = None
        self._led_task = None
        self.serial_connections = set()         # type: Set[FastSerialCommunicator]
        self.fast_leds = {}
        self.flag_led_tick_registered = False
        self.hw_switch_data = None
        self.io_boards = {}     # type: Dict[int, FastIoBoard]

        self.fast_commands = {'ID': lambda x, y: None,  # processor ID
                              'WX': lambda x, y: None,  # watchdog
                              'NI': lambda x, y: None,  # node ID
                              'RX': lambda x, y: None,  # RGB cmd received
                              'RA': lambda x, y: None,  # RGB all cmd received
                              'RF': lambda x, y: None,  # RGB fade cmd received
                              'DX': lambda x, y: None,  # DMD cmd received
                              'SX': lambda x, y: None,  # sw config received
                              'LX': lambda x, y: None,  # lamp cmd received
                              'PX': lambda x, y: None,  # segment cmd received
                              'WD': lambda x, y: None,  # watchdog
                              'SA': self.receive_sa,  # all switch states
                              '/N': self.receive_nw_open,    # nw switch open
                              '-N': self.receive_nw_closed,  # nw switch closed
                              '/L': self.receive_local_open,    # local sw open
                              '-L': self.receive_local_closed,  # local sw cls
                              '!B': self.receive_bootloader,    # nano bootloader message
                              }

    def get_info_string(self):
        """Dump infos about boards."""
        infos = ""
        if not self.net_connection:
            infos += "No connection to the NET CPU.\n"
        else:
            infos += "NET CPU: {} {} {}\n".format(
                self.net_connection.remote_processor,
                self.net_connection.remote_model,
                self.net_connection.remote_firmware)
        if not self.rgb_connection:
            infos += "No connection to the RGB CPU.\n"
        else:
            infos += "RGB CPU: {} {} {}\n".format(
                self.rgb_connection.remote_processor,
                self.rgb_connection.remote_model,
                self.rgb_connection.remote_firmware)
        if not self.dmd_connection:
            infos += "No connection to the DMD CPU.\n"
        else:
            infos += "DMD CPU: {} {} {}\n".format(
                self.dmd_connection.remote_processor,
                self.dmd_connection.remote_model,
                self.dmd_connection.remote_firmware)

        infos += "\nBoards:\n"
        for board in self.io_boards.values():
            infos += board.get_description_string() + "\n"
        return infos

    def _update_net(self) -> str:
        """Update NET CPU."""
        infos = ""
        if not self.net_connection:
            infos += "No NET CPU connected. Cannot update.\n"
            return infos
        infos += "NET CPU is version {}\n".format(self.net_connection.remote_firmware)
        max_firmware = self.net_connection.remote_firmware
        update_config = None
        for update in self.config['firmware_updates']:
            if StrictVersion(update['version']) > StrictVersion(max_firmware) and update['type'] == "net":
                update_config = update

        if not update_config:
            infos += "Firmware is up to date. Will not update.\n"
            return infos
        infos += "Found an update to version {} for the NET CPU. Will flash file {}\n".format(
            update_config['version'], update_config['file'])
        firmware_file = os.path.join(self.machine.machine_path, update_config['file'])
        try:
            with open(firmware_file) as f:
                update_string = f.read().replace("\n", "\r")
        except FileNotFoundError:
            infos += "Could not find update file.\b"
            return infos
        self.net_connection.writer.write(update_string.encode())
        infos += "Update done.\n"
        return infos

    def update_firmware(self) -> str:
        """Upgrade the firmware of the CPUs."""
        return self._update_net()

    async def initialize(self):
        """Initialise platform."""
        await self._connect_to_hardware()

    def stop(self):
        """Stop platform and close connections."""
        if self._led_task:
            self._led_task.cancel()
            self._led_task = None
        if self._watchdog_task:
            self._watchdog_task.cancel()
            self._watchdog_task = None
        if self.net_connection:
            # set watchdog to expire in 1ms
            self.net_connection.writer.write(b'WD:1\r')
        if self.rgb_connection:
            self.rgb_connection.writer.write(b'BL:AA55\r')  # reset CPU using bootloader
        if self.dmd_connection:
            self.dmd_connection.writer.write(b'BL:AA55\r')  # reset CPU using bootloader

        # wait 100ms for the messages to be send
        self.machine.clock.loop.run_until_complete(asyncio.sleep(.1))

        if self.net_connection:
            self.net_connection.stop()
            self.net_connection = None

        if self.rgb_connection:
            self.rgb_connection.stop()
            self.rgb_connection = None

        if self.dmd_connection:
            self.dmd_connection.stop()
            self.dmd_connection = None

        self.serial_connections = set()

    async def start(self):
        """Start listening for commands and schedule watchdog."""
        self._watchdog_task = self.machine.clock.schedule_interval(self._update_watchdog,
                                                                   self.config['watchdog'] / 2000)

        for connection in self.serial_connections:
            await connection.start_read_loop()

    def __repr__(self):
        """Return str representation."""
        return '<Platform.FAST>'

    def register_io_board(self, board):
        """Register an IO board.

        Args:
        ----
            board: 'mpf.platform.fast.fast_io_board.FastIoBoard' to register
        """
        if board.node_id in self.io_boards:
            raise AssertionError("Duplicate node_id")
        self.io_boards[board.node_id] = board

    def _update_watchdog(self):
        """Send Watchdog command."""
        self.net_connection.send('WD:' + str(hex(self.config['watchdog']))[2:])

    def process_received_message(self, msg: str, remote_processor: str):
        """Send an incoming message from the FAST controller to the proper method for servicing.

        Args:
        ----
            msg: messaged which was received
            remote_processor: Processor which sent the message.
        """
        assert self.log is not None
        if msg == "!SRE":
            # ignore system interrupt
            self.log.info("Received system interrupt from %s.", remote_processor)
            return

        if msg[2:3] == ':':
            cmd = msg[0:2]
            payload = msg[3:].replace('\r', '')
        else:   # pragma: no cover
            self.log.warning("Received malformed message: %s from %s", msg, remote_processor)
            return

        # Can't use try since it swallows too many errors for now
        if cmd in self.fast_commands:
            self.fast_commands[cmd](payload, remote_processor)
        else:   # pragma: no cover
            self.log.warning("Received unknown serial command? %s from %s.", msg, remote_processor)

    async def _connect_to_hardware(self):
        """Connect to each port from the config.

        This process will cause the connection threads to figure out which processor they've connected to
        and to register themselves.
        """
        ports = None
        if self.config['ports'][0] == "autodetect":
            devices = [port.device for port in serial.tools.list_ports.comports()]
            # Look for four devices with sequential tails of 0-3 or A-D
            seqs = (("0", "1", "2", "3"), ("A", "B", "C", "D"))
            for d in devices:
                for seq in seqs:
                    if d[-1] == seq[0]:
                        root = d[:-1]
                        if "{}{}".format(root, seq[1]) in devices and \
                           "{}{}".format(root, seq[2]) in devices and \
                           "{}{}".format(root, seq[3]) in devices:
                            ports = ("{}{}".format(root, seq[1]), "{}{}".format(root, seq[2]))
                            self.debug_log("Autodetect found ports! {}".format(ports))
                            break
                # If ports were found, skip the rest of the devices
                if ports:
                    break
            if not ports:
                raise RuntimeError("Unable to auto-detect FAST hardware from available devices: {}".format(
                                   ", ".join(devices)))
        else:
            ports = self.config['ports']

        for port in ports:
            comm = FastSerialCommunicator(platform=self, port=port,
                                          baud=self.config['baud'])
            await comm.connect()
            self.serial_connections.add(comm)

    def register_processor_connection(self, name: str, communicator):
        """Register processor.

        Once a communication link has been established with one of the
        processors on the FAST board, this method lets the communicator let MPF
        know which processor it's talking to.

        This is a separate method since we don't know which processor is on
        which serial port ahead of time.

        Args:
        ----
            communicator: communicator object
            name: name of processor
        """
        if name == 'DMD':
            self.dmd_connection = communicator
        elif name == 'NET':
            self.net_connection = communicator
        elif name == 'RGB':
            self.rgb_connection = communicator
            self.rgb_connection.send('RF:0')
            self.rgb_connection.send('RA:000000')  # turn off all LEDs
            self.rgb_connection.send('RF:{}'.format(
                Util.int_to_hex_string(self.config['hardware_led_fade_time'])))

    def update_leds(self):
        """Update all the LEDs connected to a FAST controller.

        This is done once per game loop for efficiency (i.e. all LEDs are sent as a single
        update rather than lots of individual ones).

        Also, every LED is updated every loop, even if it doesn't change. This
        is in case some interference causes a LED to change color. Since we
        update every loop, it will only be the wrong color for one tick.
        """
        dirty_leds = [led for led in self.fast_leds.values() if led.dirty]

        if dirty_leds:
            msg = 'RS:' + ','.join(["%s%s" % (led.number, led.current_color) for led in dirty_leds])
            self.rgb_connection.send(msg)

    async def get_hw_switch_states(self):
        """Return hardware states."""
        return self.hw_switch_data

    def receive_nw_open(self, msg, remote_processor):
        """Process network switch open.

        Args:
        ----
            msg: switch number
            remote_processor: Processor which sent the message.
        """
        assert remote_processor == "NET"
        self.machine.switch_controller.process_switch_by_num(state=0,
                                                             num=(msg, 1),
                                                             platform=self)

    def receive_nw_closed(self, msg, remote_processor):
        """Process network switch closed.

        Args:
        ----
            msg: switch number
            remote_processor: Processor which sent the message.
        """
        assert remote_processor == "NET"
        self.machine.switch_controller.process_switch_by_num(state=1,
                                                             num=(msg, 1),
                                                             platform=self)

    def receive_local_open(self, msg, remote_processor):
        """Process local switch open.

        Args:
        ----
            msg: switch number
            remote_processor: Processor which sent the message.
        """
        assert remote_processor == "NET"
        self.machine.switch_controller.process_switch_by_num(state=0,
                                                             num=(msg, 0),
                                                             platform=self)

    def receive_local_closed(self, msg, remote_processor):
        """Process local switch closed.

        Args:
        ----
            msg: switch number
            remote_processor: Processor which sent the message.
        """
        assert remote_processor == "NET"
        self.machine.switch_controller.process_switch_by_num(state=1,
                                                             num=(msg, 0),
                                                             platform=self)

    def receive_sa(self, msg, remote_processor):
        """Receive all switch states.

        Args:
        ----
            msg: switch states as bytearray
            remote_processor: Processor which sent the message.
        """
        assert remote_processor == "NET"
        self.debug_log("Received SA: %s", msg)

        hw_states = dict()

        _, local_states, _, nw_states = msg.split(',')

        for offset, byte in enumerate(bytearray.fromhex(nw_states)):
            for i in range(8):
                num = Util.int_to_hex_string((offset * 8) + i)
                if byte & (2**i):
                    hw_states[(num, 1)] = 1
                else:
                    hw_states[(num, 1)] = 0

        for offset, byte in enumerate(bytearray.fromhex(local_states)):
            for i in range(8):

                num = Util.int_to_hex_string((offset * 8) + i)

                if byte & (2**i):
                    hw_states[(num, 0)] = 1
                else:
                    hw_states[(num, 0)] = 0

        self.hw_switch_data = hw_states

    @staticmethod
    def convert_number_from_config(number):
        """Convert a number from config format to hex."""
        return Util.int_to_hex_string(number)

    def _parse_driver_number(self, number):
        try:
            board_str, driver_str = number.split("-")
        except ValueError:
            total_drivers = 0
            for board_obj in self.io_boards.values():
                total_drivers += board_obj.driver_count
            index = self.convert_number_from_config(number)

            if int(index, 16) >= total_drivers:
                raise AssertionError("Driver {} does not exist. Only {} drivers found. Driver number: {}".format(
                    int(index, 16), total_drivers, number))

            return index

        board = int(board_str)
        driver = int(driver_str)

        if board not in self.io_boards:
            raise AssertionError("Board {} does not exist for driver {}".format(board, number))

        if self.io_boards[board].driver_count <= driver:
            raise AssertionError("Board {} only has {} drivers. Driver: {}".format(
                board, self.io_boards[board].driver_count, number))

        index = 0
        for board_number, board_obj in self.io_boards.items():
            if board_number >= board:
                continue
            index += board_obj.driver_count

        return Util.int_to_hex_string(index + driver)

    def configure_driver(self, config: DriverConfig, number: str, platform_settings: dict) -> FASTDriver:
        """Configure a driver.

        Args:
        ----
            config: Driver config.
            number: Number of this driver.
            platform_settings: Platform specific settings.

        Returns: Driver object
        """
        # dont modify the config. make a copy
        platform_settings = deepcopy(platform_settings)

        if not self.net_connection:
            raise AssertionError('A request was made to configure a FAST '
                                 'driver, but no connection to a NET processor'
                                 'is available')

        if not number:
            raise AssertionError("Driver needs a number")

        # If we have WPC driver boards, look up the driver number
        if self.machine_type == 'wpc':
            try:
                number = fast_defines.WPC_DRIVER_MAP[number.upper()]
            except KeyError:
                self.raise_config_error("Could not find WPC driver {}".format(number), 1)

            if ('connection' in platform_settings and
                    platform_settings['connection'].lower() == 'network'):
                platform_settings['connection'] = 1
            else:
                platform_settings['connection'] = 0  # local driver (default for WPC)

        # If we have FAST IO boards, we need to make sure we have hex strings
        elif self.machine_type == 'fast':

            number = self._parse_driver_number(number)

            # Now figure out the connection type
            if ('connection' in platform_settings and
                    platform_settings['connection'].lower() == 'local'):
                platform_settings['connection'] = 0
            else:
                platform_settings['connection'] = 1  # network driver (default for FAST)

        else:
            raise AssertionError("Invalid machine type: {}".format(
                self.machine_type))

        return FASTDriver(config, self, number, platform_settings)

    def _parse_servo_number(self, number):
        try:
            board_str, servo_str = number.split("-")
        except ValueError:
            return self.convert_number_from_config(number)

        board = int(board_str)
        servo = int(servo_str)
        if board < 0:
            raise AssertionError("Board needs to be positive.")

        if servo < 0 or servo > 5:
            raise AssertionError("Servo number has to be between 0 and 5.")

        # every servo board supports exactly 6 servos
        return self.convert_number_from_config(board * 6 + servo)

    async def configure_servo(self, number: str) -> FastServo:
        """Configure a servo.

        Args:
        ----
            number: Number of servo

        Returns: Servo object.
        """
        number_int = self._parse_servo_number(str(number))

        return FastServo(number_int, self.net_connection)

    def _parse_switch_number(self, number):
        try:
            board_str, switch_str = number.split("-")
        except ValueError:
            total_switches = 0
            for board_obj in self.io_boards.values():
                total_switches += board_obj.switch_count
            index = self.convert_number_from_config(number)

            if int(index, 16) >= total_switches:
                raise AssertionError("Switch {} does not exist. Only {} switches found. Switch number: {}".format(
                    int(index, 16), total_switches, number))

            return index

        board = int(board_str)
        switch = int(switch_str)

        if board not in self.io_boards:
            raise AssertionError("Board {} does not exist for switch {}".format(board, number))

        if self.io_boards[board].switch_count <= switch:
            raise AssertionError("Board {} only has {} switches. Switch: {}".format(
                board, self.io_boards[board].switch_count, number))

        index = 0
        for board_number, board_obj in self.io_boards.items():
            if board_number >= board:
                continue
            index += board_obj.switch_count

        return Util.int_to_hex_string(index + switch)

    def configure_switch(self, number: str, config: SwitchConfig, platform_config: dict) -> FASTSwitch:
        """Configure the switch object for a FAST Pinball controller.

        FAST Controllers support two types of switches: `local` and `network`.
        Local switches are switches that are connected to the FAST controller
        board itself, and network switches are those connected to a FAST I/O
        board.

        MPF needs to know which type of switch is this is. You can specify the
        switch's connection type in the config file via the ``connection:``
        setting (either ``local`` or ``network``).

        If a connection type is not specified, this method will use some
        intelligence to try to figure out which default should be used.

        If the DriverBoard type is ``fast``, then it assumes the default is
        ``network``. If it's anything else (``wpc``, ``system11``, ``bally``,
        etc.) then it assumes the connection type is ``local``. Connection types
        can be mixed and matched in the same machine.

        Args:
        ----
            number: Number of this switch.
            config: Switch config.
            platform_config: Platform specific settings.

        Returns: Switch object.
        """
        if not number:
            raise AssertionError("Switch needs a number")

        if not self.net_connection:
            raise AssertionError("A request was made to configure a FAST "
                                 "switch, but no connection to a NET processor"
                                 "is available")

        if self.machine_type == 'wpc':  # translate switch num to FAST switch
            try:
                number = fast_defines.WPC_SWITCH_MAP[str(number).upper()]
            except KeyError:
                self.raise_config_error("Could not find WPC switch {}".format(number), 2)

            if 'connection' not in platform_config:
                platform_config['connection'] = 0  # local switch (default for WPC)
            else:
                platform_config['connection'] = 1  # network switch

        elif self.machine_type == 'fast':
            if 'connection' not in platform_config:
                platform_config['connection'] = 1  # network switch (default for FAST)
            else:
                platform_config['connection'] = 0  # local switch

            try:
                number = self._parse_switch_number(number)
            except ValueError:
                raise AssertionError("Could not parse switch number {}/{}. Seems "
                                     "to be not a valid switch number for the"
                                     "FAST platform.".format(config.name, number))

        # convert the switch number into a tuple which is:
        # (switch number, connection)
        number_tuple = (number, platform_config['connection'])

        self.debug_log("FAST Switch hardware tuple: %s (%s)", number, config.name)

        switch = FASTSwitch(config=config, number_tuple=number_tuple,
                            platform=self, platform_settings=platform_config)

        return switch

    def configure_light(self, number, subtype, config, platform_settings) -> LightPlatformInterface:
        """Configure light in platform."""
        del config
        if not self.net_connection:
            raise AssertionError('A request was made to configure a FAST Light, '
                                 'but no connection to a NET processor is '
                                 'available')
        if subtype == "gi":
            return FASTGIString(number, self.net_connection.send, self.machine,
                                int(1 / self.machine.config['mpf']['default_light_hw_update_hz'] * 1000))
        if subtype == "matrix":
            return FASTMatrixLight(number, self.net_connection.send, self.machine,
                                   int(1 / self.machine.config['mpf']['default_light_hw_update_hz'] * 1000), self)
        if not subtype or subtype == "led":
            if not self.flag_led_tick_registered:
                # Update leds every frame
                self._led_task = self.machine.clock.schedule_interval(
                    self.update_leds, 1 / self.machine.config['mpf']['default_light_hw_update_hz'])
                self.flag_led_tick_registered = True

            number_str, channel = number.split("-")
            if number_str not in self.fast_leds:
                self.fast_leds[number_str] = FASTDirectLED(
                    number_str, int(self.config['hardware_led_fade_time']), self.machine)
            fast_led_channel = FASTDirectLEDChannel(self.fast_leds[number_str], channel)
            self.fast_leds[number_str].add_channel(int(channel), fast_led_channel)

            return fast_led_channel

        raise AssertionError("Unknown subtype {}".format(subtype))

    def parse_light_number_to_channels(self, number: str, subtype: str):
        """Parse light channels from number string."""
        if subtype == "gi":
            if self.machine_type == 'wpc':  # translate number to FAST GI number
                try:
                    number = fast_defines.WPC_GI_MAP[str(number).upper()]
                except KeyError:
                    self.raise_config_error("Could not find WPC GI {}".format(number), 3)
            else:
                number = self.convert_number_from_config(number)

            return [
                {
                    "number": number
                }
            ]
        if subtype == "matrix":
            if self.machine_type == 'wpc':  # translate number to FAST light num
                try:
                    number = fast_defines.WPC_LIGHT_MAP[str(number).upper()]
                except KeyError:
                    self.raise_config_error("Could not find WPC light {}".format(number), 4)
            else:
                number = self.convert_number_from_config(number)

            return [
                {
                    "number": number
                }
            ]
        if not subtype or subtype == "led":
            # if the LED number is in <channel> - <led> format, convert it to a
            # FAST hardware number
            if '-' in str(number):
                num = str(number).split('-')
                index = (int(num[0]) * 64) + int(num[1])
            else:
                index = int(number)
            return [
                {
                    "number": "{}-0".format(index)
                },
                {
                    "number": "{}-1".format(index)
                },
                {
                    "number": "{}-2".format(index)
                },
            ]

        raise AssertionError("Unknown subtype {}".format(subtype))

    def configure_dmd(self):
        """Configure a hardware DMD connected to a FAST controller."""
        if not self.dmd_connection:
            raise AssertionError("A request was made to configure a FAST DMD, "
                                 "but no connection to a DMD processor is "
                                 "available.")

        return FASTDMD(self.machine, self.dmd_connection.send)

    @classmethod
    def get_coil_config_section(cls):
        """Return coil config section."""
        return "fast_coils"

    @classmethod
    def get_switch_config_section(cls):
        """Return switch config section."""
        return "fast_switches"

    def _check_switch_coil_combincation(self, switch, coil):
        switch_number = int(switch.hw_switch.number[0], 16)
        coil_number = int(coil.hw_driver.number, 16)

        # first 8 switches always work
        if 0 <= switch_number <= 7:
            return

        switch_index = 0
        coil_index = 0
        for board_obj in self.io_boards.values():
            # if switch and coil are on the same board we are fine
            if switch_index <= switch_number < switch_index + board_obj.switch_count and \
                    coil_index <= coil_number < coil_index + board_obj.driver_count:
                return
            coil_index += board_obj.driver_count
            switch_index += board_obj.switch_count

        raise AssertionError("Driver {} and switch {} are on different boards. Cannot apply rule!".format(
            coil.hw_driver.number, switch.hw_switch.number))

    def set_pulse_on_hit_and_release_rule(self, enable_switch, coil):
        """Set pulse on hit and release rule to driver."""
        self.debug_log("Setting Pulse on hit and release HW Rule. Switch: %s,"
                       "Driver: %s", enable_switch.hw_switch.number,
                       coil.hw_driver.number)

        self._check_switch_coil_combincation(enable_switch, coil)

        driver = coil.hw_driver

        cmd = '{}{},{},{},18,{},{},00,{},00'.format(
            driver.get_config_cmd(),
            coil.hw_driver.number,
            driver.get_control_for_cmd(enable_switch),
            enable_switch.hw_switch.number[0],
            Util.int_to_hex_string(coil.pulse_settings.duration),
            driver.get_pwm_for_cmd(coil.pulse_settings.power),
            driver.get_recycle_ms_for_cmd(coil.recycle, coil.pulse_settings.duration))

        enable_switch.hw_switch.configure_debounce(enable_switch.debounce)
        driver.set_autofire(cmd, coil.pulse_settings.duration, coil.pulse_settings.power, 0)

    def set_pulse_on_hit_and_release_and_disable_rule(self, enable_switch: SwitchSettings,
                                                      eos_switch: SwitchSettings, coil: DriverSettings,
                                                      repulse_settings: Optional[RepulseSettings]):
        """Set pulse on hit and release and disable rule on driver."""
        # Potential command from Dave:
        # Command
        # [DL/DN]:<DRIVER_ID>,<CONTROL>,<SWITCH_ID_ON>,<75>,<SWITCH_ID_OFF>,<Driver On Time1>,<Driver On Time2 X 100mS>,
        # <PWM2><Driver Rest Time><CR>#
        # SWITCH_ID_ON would be the flipper switch
        # SWITCH_ID_OFF would be the EOS switch.
        # So for the flipper, Driver On Time1 will = the maximum time the coil can be held on if the EOS fails.
        # Driver On Time2 X 100mS would not be used for a flipper, so set it to 0.
        # And PWM2 should be left on full 0xff unless you need less power for some reason.
        self.debug_log("Setting Pulse on hit and release with HW Rule. Switch:"
                       "%s, Driver: %s", enable_switch.hw_switch.number,
                       coil.hw_driver.number)

        self._check_switch_coil_combincation(enable_switch, coil)
        self._check_switch_coil_combincation(eos_switch, coil)

        driver = coil.hw_driver

        cmd = '{}{},{},{},75,{},{},00,{},{}'.format(
            driver.get_config_cmd(),
            coil.hw_driver.number,
            driver.get_control_for_cmd(enable_switch, eos_switch),
            enable_switch.hw_switch.number[0],
            eos_switch.hw_switch.number[0],
            Util.int_to_hex_string(coil.pulse_settings.duration),
            driver.get_pwm_for_cmd(coil.pulse_settings.power),
            driver.get_recycle_ms_for_cmd(coil.recycle, coil.pulse_settings.duration))

        enable_switch.hw_switch.configure_debounce(enable_switch.debounce)
        eos_switch.hw_switch.configure_debounce(eos_switch.debounce)
        driver.set_autofire(cmd, coil.pulse_settings.duration, coil.pulse_settings.power, 0)

    def set_pulse_on_hit_and_enable_and_release_and_disable_rule(self, enable_switch: SwitchSettings,
                                                                 eos_switch: SwitchSettings, coil: DriverSettings,
                                                                 repulse_settings: Optional[RepulseSettings]):
        """Set pulse on hit and enable and release and disable rule on driver."""
        self.warning_log("EOS cut-off rule will not work with FAST on single-wound coils. %s %s %s", enable_switch,
                         eos_switch, coil)
        self.set_pulse_on_hit_and_enable_and_release_rule(enable_switch, coil)

    def set_pulse_on_hit_rule(self, enable_switch: SwitchSettings, coil: DriverSettings):
        """Set pulse on hit rule on driver."""
        self.debug_log("Setting Pulse on hit and release HW Rule. Switch: %s,"
                       "Driver: %s", enable_switch.hw_switch.number,
                       coil.hw_driver.number)

        self._check_switch_coil_combincation(enable_switch, coil)

        driver = coil.hw_driver

        cmd = '{}{},{},{},10,{},{},00,00,{}'.format(
            driver.get_config_cmd(),
            coil.hw_driver.number,
            driver.get_control_for_cmd(enable_switch),
            enable_switch.hw_switch.number[0],
            Util.int_to_hex_string(coil.pulse_settings.duration),
            driver.get_pwm_for_cmd(coil.pulse_settings.power),
            driver.get_recycle_ms_for_cmd(coil.recycle, coil.pulse_settings.duration))

        enable_switch.hw_switch.configure_debounce(enable_switch.debounce)
        driver.set_autofire(cmd, coil.pulse_settings.duration, coil.pulse_settings.power, 0)

    def set_pulse_on_hit_and_enable_and_release_rule(self, enable_switch: SwitchSettings, coil: DriverSettings):
        """Set pulse on hit and enable and relase rule on driver."""
        self.debug_log("Setting Pulse on hit and enable and release HW Rule. "
                       "Switch: %s, Driver: %s",
                       enable_switch.hw_switch.number, coil.hw_driver.number)

        self._check_switch_coil_combincation(enable_switch, coil)

        driver = coil.hw_driver

        cmd = '{}{},{},{},18,{},{},{},{},00'.format(
            driver.get_config_cmd(),
            driver.number,
            driver.get_control_for_cmd(enable_switch),
            enable_switch.hw_switch.number[0],
            Util.int_to_hex_string(coil.pulse_settings.duration),
            driver.get_pwm_for_cmd(coil.pulse_settings.power),
            driver.get_hold_pwm_for_cmd(coil.hold_settings.power),
            driver.get_recycle_ms_for_cmd(coil.recycle, coil.pulse_settings.duration))

        enable_switch.hw_switch.configure_debounce(enable_switch.debounce)
        driver.set_autofire(cmd, coil.pulse_settings.duration, coil.pulse_settings.power, coil.hold_settings.power)

    def clear_hw_rule(self, switch, coil):
        """Clear a hardware rule.

        This is used if you want to remove the linkage between a switch and
        some driver activity. For example, if you wanted to disable your
        flippers (so that a player pushing the flipper buttons wouldn't cause
        the flippers to flip), you'd call this method with your flipper button
        as the *sw_num*.

        Args:
        ----
            switch: The switch whose rule you want to clear.
            coil: The coil whose rule you want to clear.

        """
        self.debug_log("Clearing HW Rule for switch: %s, coils: %s",
                       switch.hw_switch.number, coil.hw_driver.number)

        # TODO: check that the rule is switch + coil and not another switch + this coil

        driver = coil.hw_driver

        driver.clear_autofire(driver.get_config_cmd(), driver.number)

    def receive_bootloader(self, msg, remote_processor):
        """Process bootloader message."""
        self.debug_log("Got Bootloader message: %s from", msg, remote_processor)
        ignore_rgb = self.config['ignore_rgb_crash'] and \
            remote_processor == self.rgb_connection.remote_processor
        if msg in ('00', '02'):
            action = "Ignoring RGB crash and continuing play." if ignore_rgb else "MPF will exit now."
            self.error_log("The FAST %s processor rebooted. Unfortunately, that means that it lost all its state "
                           "(such as hardware rules or switch configs). This is likely caused by an unstable "
                           "power supply but it might also be a firmware bug. %s", remote_processor, action)
            if ignore_rgb:
                self.machine.events.post("fast_rgb_rebooted", msg=msg)
                return
            self.machine.stop("FAST {} rebooted during game".format(remote_processor))



from mpf.tests.MpfTestCase import MpfTestCase


class TestSmartVirtualPlatform(MpfTestCase):

    def get_config_file(self):
        return "test_virtual.yaml"

    def get_machine_path(self):
        return 'tests/machine_files/platform/'

    def test_load_config_and_allow_enable(self):
        # test that we can load the config with coil and switch parameters from platforms

        self.machine.coils["c_test_allow_enable"].enable()
        self.machine.coils["c_test_hold_power"].enable()

        with self.assertRaises(AssertionError):
            self.machine.coils["c_test_no_allow_enable"].enable()



import mpf.tests.test_BallDeviceSwitchConfirmation


class TestBallDeviceEventConfirmation(
    mpf.tests.test_BallDeviceSwitchConfirmation
        .TestBallDeviceSwitchConfirmation):

    def get_config_file(self):
        return 'test_ball_device_event_confirmation.yaml'

    def get_machine_path(self):
        return 'tests/machine_files/ball_device/'

    def _hit_confirm(self):
        self.machine.events.post("launcher_confirm")



# Generated by the gRPC Python protocol compiler plugin. DO NOT EDIT!
"""Client and server classes corresponding to protobuf-defined services."""
import grpc

from . import platform_pb2 as platform__pb2


class MpfHardwareServiceStub(object):
    """Missing associated documentation comment in .proto file."""

    def __init__(self, channel):
        """Constructor.

        Args:
            channel: A grpc.Channel.
        """
        self.Start = channel.unary_stream(
                '/mpf.vpe.MpfHardwareService/Start',
                request_serializer=platform__pb2.MachineState.SerializeToString,
                response_deserializer=platform__pb2.Commands.FromString,
                )
        self.GetMachineDescription = channel.unary_unary(
                '/mpf.vpe.MpfHardwareService/GetMachineDescription',
                request_serializer=platform__pb2.EmptyRequest.SerializeToString,
                response_deserializer=platform__pb2.MachineDescription.FromString,
                )
        self.SendSwitchChanges = channel.stream_unary(
                '/mpf.vpe.MpfHardwareService/SendSwitchChanges',
                request_serializer=platform__pb2.SwitchChanges.SerializeToString,
                response_deserializer=platform__pb2.EmptyResponse.FromString,
                )
        self.Quit = channel.unary_unary(
                '/mpf.vpe.MpfHardwareService/Quit',
                request_serializer=platform__pb2.QuitRequest.SerializeToString,
                response_deserializer=platform__pb2.EmptyResponse.FromString,
                )


class MpfHardwareServiceServicer(object):
    """Missing associated documentation comment in .proto file."""

    def Start(self, request, context):
        """Missing associated documentation comment in .proto file."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def GetMachineDescription(self, request, context):
        """Missing associated documentation comment in .proto file."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def SendSwitchChanges(self, request_iterator, context):
        """Missing associated documentation comment in .proto file."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def Quit(self, request, context):
        """Missing associated documentation comment in .proto file."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')


def add_MpfHardwareServiceServicer_to_server(servicer, server):
    rpc_method_handlers = {
            'Start': grpc.unary_stream_rpc_method_handler(
                    servicer.Start,
                    request_deserializer=platform__pb2.MachineState.FromString,
                    response_serializer=platform__pb2.Commands.SerializeToString,
            ),
            'GetMachineDescription': grpc.unary_unary_rpc_method_handler(
                    servicer.GetMachineDescription,
                    request_deserializer=platform__pb2.EmptyRequest.FromString,
                    response_serializer=platform__pb2.MachineDescription.SerializeToString,
            ),
            'SendSwitchChanges': grpc.stream_unary_rpc_method_handler(
                    servicer.SendSwitchChanges,
                    request_deserializer=platform__pb2.SwitchChanges.FromString,
                    response_serializer=platform__pb2.EmptyResponse.SerializeToString,
            ),
            'Quit': grpc.unary_unary_rpc_method_handler(
                    servicer.Quit,
                    request_deserializer=platform__pb2.QuitRequest.FromString,
                    response_serializer=platform__pb2.EmptyResponse.SerializeToString,
            ),
    }
    generic_handler = grpc.method_handlers_generic_handler(
            'mpf.vpe.MpfHardwareService', rpc_method_handlers)
    server.add_generic_rpc_handlers((generic_handler,))


 # This class is part of an EXPERIMENTAL API.
class MpfHardwareService(object):
    """Missing associated documentation comment in .proto file."""

    @staticmethod
    def Start(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_stream(request, target, '/mpf.vpe.MpfHardwareService/Start',
            platform__pb2.MachineState.SerializeToString,
            platform__pb2.Commands.FromString,
            options, channel_credentials,
            insecure, call_credentials, compression, wait_for_ready, timeout, metadata)

    @staticmethod
    def GetMachineDescription(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(request, target, '/mpf.vpe.MpfHardwareService/GetMachineDescription',
            platform__pb2.EmptyRequest.SerializeToString,
            platform__pb2.MachineDescription.FromString,
            options, channel_credentials,
            insecure, call_credentials, compression, wait_for_ready, timeout, metadata)

    @staticmethod
    def SendSwitchChanges(request_iterator,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.stream_unary(request_iterator, target, '/mpf.vpe.MpfHardwareService/SendSwitchChanges',
            platform__pb2.SwitchChanges.SerializeToString,
            platform__pb2.EmptyResponse.FromString,
            options, channel_credentials,
            insecure, call_credentials, compression, wait_for_ready, timeout, metadata)

    @staticmethod
    def Quit(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(request, target, '/mpf.vpe.MpfHardwareService/Quit',
            platform__pb2.QuitRequest.SerializeToString,
            platform__pb2.EmptyResponse.FromString,
            options, channel_credentials,
            insecure, call_credentials, compression, wait_for_ready, timeout, metadata)



"""IRC Chat Bot for monitoring a Twitch chatroom."""
from functools import partial

import logging
import textwrap

try:
    from irc.bot import SingleServerIRCBot
    IMPORT_SUCCESS = True
except ImportError:
    SingleServerIRCBot = object     # prevent class loading error
    IMPORT_SUCCESS = False

MYPY = False
if MYPY:   # pragma: no cover
    import asyncio  # pylint: disable-msg=cyclic-import,unused-import


class TwitchClient(SingleServerIRCBot):

    """Thread to process Twitch chat events."""

    # pylint: disable-msg=too-many-arguments
    def __init__(self, machine, username, password, channel, loop):
        """Initialize Twitch Bot."""
        self.log = logging.getLogger('twitch_client')
        self.machine = machine
        self.password = password
        self.channel = '#' + channel.lower()
        self.loop = loop    # type: asyncio.AbstractEventLoop

        if not IMPORT_SUCCESS:
            raise AssertionError('Please install irc extension using "pip3 install irc" to use the twitch client.')

        # Create IRC bot connection
        server = 'irc.chat.twitch.tv'
        port = 6667
        self.log.info('Connecting to %s on port %s...', server, port)
        super().__init__([(
            server, port,
            (password if password.lower().startswith('oauth:') else ('oauth:' + password))
        )], username, username)
        # self.connection.add_global_handler("all_events", self.on_all_events, -100)

    def on_welcome(self, c, e):
        """Framework will call when IRC server is joined."""
        del e
        self.log.info('Joining %s', self.channel)

        # You must request specific capabilities before you can use them
        c.cap('REQ', ':twitch.tv/membership')
        c.cap('REQ', ':twitch.tv/tags')
        c.cap('REQ', ':twitch.tv/commands')
        c.join(self.channel)

    def on_pubmsg(self, c, e):
        """Framework will call when a public message is posted in chat."""
        del c
        # If a chat message starts with ! or ?, try to run it as a command
        if e.arguments[0][:1] == '!' or e.arguments[0][:1] == '?':
            cmd = e.arguments[0].split(' ')[0][1:]
            self.do_command(e, cmd.lower())

        self.process_twitch_event(e)

    def on_privmsg(self, c, e):
        """Framework will call when a private message is posted in chat."""
        del c
        self.log.info('Private Chat: %s', e.arguments[0])

    def on_usernotice(self, c, e):
        """Framework will call when a user notice is posted in chat."""
        del c
        self.process_twitch_event(e)

    def on_all_events(self, c, e):
        """Framework will call when any IRC event is posted."""
        del c
        message = 'All Events: ' + e
        self.log.info(message.replace(self.password, 'XXXXX'))

    def do_command(self, e, cmd):
        """Handle a chat command (starts with ? or !)."""
        user = e.source.split('!')[0]
        self.log.info('Received command: [' + user + '] ' + cmd.lower())

        self.post_event_in_mpf('twitch_command', user=user, command=cmd.lower())
        '''event: twitch_command
        desc: A user typed a line that begins with ! or ?
        args:
        command: The text after the ! or ?
        user: The chat user who executed the command
        '''

    def post_event_in_mpf(self, event, *args, **kwargs):
        """Post event in MPF via async loop to prevent race conditions."""
        self.loop.call_soon_threadsafe(partial(self.machine.events.post, event, *args, **kwargs))

    def set_machine_variable_in_mpf(self, name, value):
        """Set machine var in MPF via async loop to prevent race conditions."""
        self.loop.call_soon_threadsafe(self.machine.variables.set_machine_var, name, value)

    def is_connected(self):
        """Return true if the server is connected."""
        return self.connection.is_connected()

    def process_twitch_event(self, e):
        """Process the event and delegate to the proper handler."""
        tags = self.build_tag_dict(e.tags)
        message = next(iter(e.arguments or []), '')
        bits = tags.get('bits')
        message_type = tags.get('msg-id', '')
        user = tags.get('display-name', 'Unknown')
        is_sub = message_type in ('sub', 'resub')
        is_gift_sub = message_type in ('subgift', 'anonsubgift')
        is_raid = message_type == 'raid'

        log_message = 'Chat: [' + user + '] ' + message + ' : ' + str(e)
        self.log.info(log_message.replace(self.password, 'XXXXX'))

        if is_sub or is_gift_sub:
            self.process_subscription(user, message, tags, is_gift_sub)
        elif is_raid:
            self.process_raid(user, tags)
        elif bits is not None:
            self.process_bits(user, message, bits)
        else:
            self.process_chat(user, message)

    def process_subscription(self, user, message, tags, is_gift):
        """Send the subscription event to MPF."""
        months = tags.get('msg-param-months' if is_gift else 'msg-param-cumulative-months', 1)
        subscriber_message = tags.get('message', '')
        sub_plan = tags.get('msg-param-sub-plan', '')
        sub_plan_name = tags.get('msg-param-sub-plan-name', '')
        sub_recipient = tags.get('msg-param-recipient-display-name', user) if is_gift else user
        self.set_machine_variable_in_mpf('twitch_last_sub_user', user)
        self.set_machine_variable_in_mpf('twitch_last_sub_recipient', sub_recipient)
        self.set_machine_variable_in_mpf('twitch_last_sub_months', int(months))
        self.set_machine_variable_in_mpf('twitch_last_sub_message', subscriber_message)
        self.set_machine_variable_in_mpf('twitch_last_sub_plan', sub_plan)
        self.set_machine_variable_in_mpf('twitch_last_sub_plan_name', sub_plan_name)
        self.set_machine_variable_in_mpf('twitch_last_sub_is_gift', is_gift)
        self.post_event_in_mpf(
            'twitch_subscription',
            user=user,
            message=message,
            months=int(months),
            subscriber_message=subscriber_message,
            sub_plan=sub_plan,
            sub_plan_name=sub_plan_name,
            sub_recipient=sub_recipient,
            gift=is_gift
        )
        '''event: twitch_subscription
        desc: A chat user has subscribed or resubscribed on Twitch
        args:
        gift: True if this sub was gifted by another user
        message: Chat message text
        months: The number of months that the user has been a subscriber
        sub_plan: The subscription tier (Prime, 1000, 2000, 3000)
        sub_plan_name: The streamer specific name for the sub tier
        sub_recipient: The user who is subscribing
        subscriber_message: The message the user typed when subscribing
        user: The chat user name who paid for the subscription
        '''

    def process_raid(self, user, tags):
        """Send the raid event to MPF."""
        raid_user = tags.get('msg-param-displayName', user)
        raid_count = tags.get('msg-param-viewerCount', 0)
        self.set_machine_variable_in_mpf('twitch_last_raid_user', raid_user)
        self.set_machine_variable_in_mpf('twitch_last_raid_count', int(raid_count))
        self.post_event_in_mpf(
            'twitch_raid',
            raid_user=raid_user,
            raid_count=int(raid_count)
        )
        '''event: twitch_raid
        desc: Another Twitch user has raided your channel
        args:
        raid_user: The user who raided
        raid_count: The count of viewers in the raid
        '''

    def process_bits(self, user, message, bits):
        """Send the bit donation event to MPF."""
        self.set_machine_variable_in_mpf('twitch_last_bits_user', user)
        self.set_machine_variable_in_mpf('twitch_last_bits_amount', int(bits))
        self.post_event_in_mpf('twitch_bit_donation', user=user, message=message, bits=int(bits))
        '''event: twitch_bit_donation
        desc: A chat user has donated bits on Twitch
        args:
        message: Chat message text
        bits: The number of bits donated
        user: The chat user name who subscribed
        '''

    def process_chat(self, user, message):
        """Send the chat event to MPF."""
        length, lines = self.split_message(message, 6)
        self.set_machine_variable_in_mpf('twitch_last_chat_user', user)
        self.set_machine_variable_in_mpf('twitch_last_chat_message', message)
        self.set_machine_variable_in_mpf('twitch_last_chat_message_line_count', length)
        self.set_machine_variable_in_mpf('twitch_last_chat_message_line_1', lines[0])
        self.set_machine_variable_in_mpf('twitch_last_chat_message_line_2', lines[1])
        self.set_machine_variable_in_mpf('twitch_last_chat_message_line_3', lines[2])
        self.set_machine_variable_in_mpf('twitch_last_chat_message_line_4', lines[3])
        self.set_machine_variable_in_mpf('twitch_last_chat_message_line_5', lines[4])
        self.set_machine_variable_in_mpf('twitch_last_chat_message_line_6', lines[5])
        self.post_event_in_mpf(
            'twitch_chat_message',
            user=user,
            message=message,
            line_count=length,
            line_1=lines[0],
            line_2=lines[1],
            line_3=lines[2],
            line_4=lines[3],
            line_5=lines[4],
            line_6=lines[5]
        )
        '''event: twitch_chat_message
        desc: A chat message was received via Twitch
        args:
        line_count: The number of lines that the text splitter produced
        line_1: Split line 1
        line_2: Split line 2
        line_3: Split line 3
        line_4: Split line 4
        line_5: Split line 5
        line_6: Split line 6
        message: Full chat message text
        user: The chat user name who subscribed
        '''

    @staticmethod
    def build_tag_dict(seq):
        """Build a Python dict from IRC chat tags."""
        return dict((d['key'], d['value']) for (index, d) in enumerate(seq))

    @staticmethod
    def split_message(message, min_lines):
        """Split up a string into lines broken on words."""
        lines = textwrap.wrap(message, 21)
        length = len(lines)

        if length < min_lines:
            lines += [''] * (min_lines - len(lines))

        return length, lines



"""Test Light Position Mixin."""
from mpf.core.rgb_color import RGBColor
from mpf.tests.MpfFakeGameTestCase import MpfFakeGameTestCase


class TestLightPositions(MpfFakeGameTestCase):

    def get_config_file(self):
        return 'light.yaml'

    def get_machine_path(self):
        return 'tests/machine_files/light/'

    def test_light_positions(self):
        led1 = self.machine.lights["led1"]
        led2 = self.machine.lights["led2"]
        led3 = self.machine.lights["led3"]

        self.assertEqual(led1.x, 0.4)
        self.assertEqual(led1.y, 0.5)
        self.assertEqual(led1.z, 0)

        self.assertEqual(led2.x, 0.6)
        self.assertEqual(led2.y, 0.7)
        self.assertEqual(led2.z, None)

        self.assertEqual(led3.x, None)
        self.assertEqual(led3.y, None)
        self.assertEqual(led3.z, None)



"""Support for incandescent wings in OPP."""
import logging

from mpf.platforms.interfaces.light_platform_interface import LightPlatformSoftwareFade
from mpf.platforms.opp.opp_modern_lights import OPPModernLightChannel

from mpf.platforms.opp.opp_rs232_intf import OppRs232Intf


class OPPIncandCard:

    """An incandescent wing card."""

    __slots__ = ["log", "addr", "chain_serial", "old_state", "new_state", "card_num", "machine", "hardware_fade_ms",
                 "mask"]

    # pylint: disable-msg=too-many-arguments
    def __init__(self, chain_serial, addr, mask, machine):
        """Initialise OPP incandescent card."""
        self.log = logging.getLogger('OPPIncand {} on {}'.format(addr, chain_serial))
        self.addr = addr
        self.chain_serial = chain_serial
        self.old_state = None
        self.new_state = 0
        self.mask = mask
        self.machine = machine
        self.card_num = str(addr - ord(OppRs232Intf.CARD_ID_GEN2_CARD))
        self.hardware_fade_ms = int(1 / machine.config['mpf']['default_light_hw_update_hz'] * 1000)

        self.log.debug("Creating OPP Incand at hardware address: 0x%02x", addr)

    def configure_software_fade_incand(self, number):
        """Configure traditional incand."""
        return OPPIncand(self, number, self.hardware_fade_ms, self.machine.clock.loop)

    def configure_modern_fade_incand(self, number, light_system):
        """Configure modern incand with fade."""
        return OPPModernLightChannel(self.chain_serial, int(self.card_num), int(number) + 0x1000, light_system)

    def is_valid_light_number(self, number):
        """Check if incand light exists in hardware."""
        return ((1 << int(number)) & self.mask) != 0


class OPPIncand(LightPlatformSoftwareFade):

    """A driver of an incandescent wing card."""

    __slots__ = ["incand_card", "index"]

    def __init__(self, incand_card, number, hardware_fade_ms, loop):
        """Initialise Incandescent wing card driver."""
        super().__init__(number, loop, hardware_fade_ms)
        self.incand_card = incand_card  # type: OPPIncandCard
        self.index = int(number)

    def set_brightness(self, brightness: float):
        """Enable (turns on) this driver.

        Args:
        ----
            brightness: brightness 0 (off) to 255 (on) for this incandescent light. OPP only supports on (>0) or off.
        """
        curr_bit = (1 << self.index)
        if brightness == 0:
            self.incand_card.new_state &= ~curr_bit
        else:
            self.incand_card.new_state |= curr_bit

    def get_board_name(self):
        """Return OPP chain and addr."""
        return "OPP {} Board {}".format(str(self.incand_card.chain_serial), "0x%02x" % self.incand_card.addr)

    def is_successor_of(self, other):
        """Return true if the other light has the previous index and is on the same card."""
        return self.incand_card == other.incand_card and self.index == self.index + 1

    def get_successor_number(self):
        """Return next index on node."""
        return "{}-{}-{}".format(self.incand_card.chain_serial, self.incand_card.addr, self.index + 1)

    def __lt__(self, other):
        """Order lights by their order on the hardware."""
        return ((self.incand_card.chain_serial, self.incand_card.addr, self.index) <
                (other.incand_card.chain_serial, other.incand_card.addr, other.index))



"""Test the LED device."""
from mpf.core.rgb_color import RGBColor
from mpf.tests.MpfTestCase import MpfTestCase, test_config


class TestDeviceLight(MpfTestCase):

    def get_config_file(self):
        return 'light.yaml'

    def get_machine_path(self):
        return 'tests/machine_files/light/'

    def test_default_on_color(self):
        led = self.machine.lights["led1"]

        # color on should map to red
        led.color("on")
        self.assertLightColor("led1", RGBColor("red"))

        # turn off again
        led.off()
        self.assertLightColor("led1", RGBColor("black"))

        # on should also map to red
        led.on()
        self.assertLightColor("led1", RGBColor("red"))

        # on with half brightness should map to dimmed red
        led.on(127)
        self.assertLightColor("led1", RGBColor("red%50"))

    @test_config("light_default_color_correction.yaml")
    def test_default_color_correction(self):
        led = self.machine.lights["led1"]
        led.color(RGBColor("white"))
        self.advance_time_and_run()
        # color is uncorrected
        self.assertLightColor("led1", RGBColor("white"))
        # corrected color
        self.assertEqual(RGBColor([210, 184, 159]), led.color_correct(led.get_color()))
        # check hardware
        self.assertEqual(210 / 255.0, led.hw_drivers["red"][0].current_brightness)
        self.assertEqual(184 / 255.0, led.hw_drivers["green"][0].current_brightness)
        self.assertEqual(159 / 255.0, led.hw_drivers["blue"][0].current_brightness)

        led.color(RGBColor([128, 128, 128]))
        self.advance_time_and_run()
        self.assertLightColor("led1", [128, 128, 128])
        self.assertEqual(RGBColor([96, 83, 70]), led.color_correct(led.get_color()))
        self.assertEqual(96 / 255.0, led.hw_drivers["red"][0].current_brightness)
        self.assertEqual(83 / 255.0, led.hw_drivers["green"][0].current_brightness)
        self.assertEqual(70 / 255.0, led.hw_drivers["blue"][0].current_brightness)

        led.color(RGBColor("black"))
        self.advance_time_and_run()
        self.assertLightColor("led1", [0, 0, 0])
        self.assertEqual(RGBColor([0, 0, 0]), led.color_correct(led.get_color()))
        self.assertEqual(0 / 255.0, led.hw_drivers["red"][0].current_brightness)
        self.assertEqual(0 / 255.0, led.hw_drivers["green"][0].current_brightness)
        self.assertEqual(0 / 255.0, led.hw_drivers["blue"][0].current_brightness)

    def test_consecutive_fades(self):
        self.assertLightColor("led1", [0, 0, 0])
        led = self.machine.lights["led1"]
        led.color(RGBColor("red"), fade_ms=1000, key="one")
        self.advance_time_and_run(.5)
        self.assertLightColor("led1", [127, 0, 0])
        self.advance_time_and_run(1)
        self.assertLightColor("led1", "red")

        # play another color with the same key
        led.color(RGBColor("red"), fade_ms=1000, key="one")
        self.advance_time_and_run(.1)
        self.assertLightColor("led1", "red")
        self.advance_time_and_run(1)

        # remove key and play again
        led.remove_from_stack_by_key("one", fade_ms=1000)
        led.color(RGBColor("red"), fade_ms=1000, key="one")
        self.advance_time_and_run(.1)
        self.assertLightColor("led1", "red")
        self.advance_time_and_run(1)

        led.remove_from_stack_by_key("one", fade_ms=1000)
        self.assertLightColor("led1", "red")
        self.advance_time_and_run(.5)
        self.assertLightColor("led1", [128, 0, 0])
        self.advance_time_and_run(.6)
        self.assertLightColor("led1", [0, 0, 0])
        self.assertFalse(led.stack)

        led.color(RGBColor("blue"), key="lower", priority=1, fade_ms=10000)
        led.color(RGBColor("red"), key="upper", priority=2, fade_ms=1000)
        self.advance_time_and_run(.5)
        self.assertLightColor("led1", [127, 0, 0])
        self.advance_time_and_run(.51)
        self.assertLightColor("led1", "red")
        self.advance_time_and_run(2)    # lower is at 3/10
        led.remove_from_stack_by_key("upper", fade_ms=4000)
        self.assertLightColor("led1", "red")
        self.advance_time_and_run(2)    # lower is at 5/10 -> [0, 0, 127]. upper at 2/4 (50% alpha)
        self.assertLightColor("led1", [128, 0, 63])
        self.advance_time_and_run(2)    # lower is at 7/10. upper is gone
        self.assertLightColor("led1", [0, 0, 178])
        self.advance_time_and_run(3)    # lower is at 10/10. upper is gone
        self.assertLightColor("led1", [0, 0, 255])
        self.assertEqual(1, len(led.stack))

    def test_color_and_stack(self):
        led1 = self.machine.lights["led1"]

        # set led1 to red and check the color and stack
        led1.color('red')

        # need to advance time since LEDs are updated once per frame via a
        # clock.schedule_interval
        self.advance_time_and_run()
        self.assertLightColor("led1", "red")

        color_setting = led1.stack[0]
        self.assertEqual(color_setting.priority, 0)
        self.assertEqual(color_setting.start_color, RGBColor('off'))
        self.assertEqual(color_setting.dest_time, 0)
        self.assertEqual(color_setting.dest_color, RGBColor('red'))
        self.assertEqual(led1.get_color(), RGBColor('red'))
        self.assertFalse(color_setting.key)

        # test get_color()
        self.assertEqual(led1.get_color(), RGBColor('red'))

        # set to blue & test
        led1.color('blue')
        self.advance_time_and_run()
        self.assertLightColor("led1", "blue")
        color_setting = led1.stack[0]
        self.assertEqual(color_setting.priority, 0)
        # self.assertEqual(color_setting.start_color, RGBColor('red'))
        self.assertEqual(color_setting.dest_time, 0)
        self.assertEqual(color_setting.dest_color, RGBColor('blue'))
        self.assertEqual(led1.get_color(), RGBColor('blue'))
        self.assertFalse(color_setting.key)
        self.assertEqual(len(led1.stack), 1)

        # set it to green, at a higher priority, but with no key. Stack should
        # reflect the higher priority, but still be len 1 since the key is the
        # same (None)
        led1.color('green', priority=100)

        self.advance_time_and_run()
        self.assertLightColor("led1", "green")
        self.assertEqual(len(led1.stack), 1)
        color_setting = led1.stack[0]
        self.assertEqual(color_setting.priority, 100)
        # self.assertEqual(color_setting.start_color, RGBColor('blue'))
        self.assertEqual(color_setting.dest_time, 0)
        self.assertEqual(color_setting.dest_color, RGBColor('green'))
        self.assertEqual(led1.get_color(), RGBColor('green'))
        self.assertFalse(color_setting.key)

        # set led1 orange, lower priority, but with a key, so led should stay
        # green, but stack len should be 2
        led1.color('orange', key='test')

        self.advance_time_and_run()
        self.assertLightColor("led1", "green")
        self.assertEqual(len(led1.stack), 2)
        color_setting = led1.stack[0]
        self.assertEqual(color_setting.priority, 100)
        # self.assertEqual(color_setting.start_color, RGBColor('blue'))
        self.assertEqual(color_setting.dest_time, 0)
        self.assertEqual(color_setting.dest_color, RGBColor('green'))
        self.assertEqual(led1.get_color(), RGBColor('green'))
        self.assertFalse(color_setting.key)

        # remove the orange key from the stack
        led1.remove_from_stack_by_key('test')
        self.assertEqual(len(led1.stack), 1)

        # clear the stack
        led1.clear_stack()
        self.assertEqual(len(led1.stack), 0)

        # test the stack ordering with different priorities & keys
        led1.color('red', priority=200, key='red')
        self.advance_time_and_run()
        self.assertLightColor("led1", "red")

        led1.color('blue', priority=300, key='blue')
        self.advance_time_and_run()
        self.assertLightColor("led1", "blue")

        led1.color('green', priority=200, key='green')
        self.advance_time_and_run()
        self.assertLightColor("led1", "blue")

        led1.color('orange', priority=100, key='orange')
        self.advance_time_and_run()
        self.assertLightColor("led1", "blue")

        # verify the stack is right
        # order should be priority, then key, so
        # should be: blue, green, red, orange
        self.assertEqual(RGBColor('blue'), led1.stack[0].dest_color)
        self.assertEqual(RGBColor('red'), led1.stack[1].dest_color)
        self.assertEqual(RGBColor('green'), led1.stack[2].dest_color)
        self.assertEqual(RGBColor('orange'), led1.stack[3].dest_color)

        # test that a replacement key slots in properly
        led1.color('red', priority=300, key='red')
        self.advance_time_and_run()
        self.assertLightColor("led1", "red")
        self.assertEqual(RGBColor('red'), led1.stack[0].dest_color)
        self.assertEqual(RGBColor('blue'), led1.stack[1].dest_color)
        self.assertEqual(RGBColor('green'), led1.stack[2].dest_color)
        self.assertEqual(RGBColor('orange'), led1.stack[3].dest_color)

    def test_named_colors(self):
        led1 = self.machine.lights["led1"]
        led1.color('jans_red')
        self.machine_run()

        self.assertLightColor(led1.name, "jans_red")
        self.assertLightColor(led1.name, [251, 23, 42])

    def test_fades(self):
        led1 = self.machine.lights["led1"]

        led1.color('red', fade_ms=2000)
        self.machine_run()

        # check the stack before the fade starts
        color_setting = led1.stack[0]
        self.assertEqual(color_setting.priority, 0)
        self.assertEqual(color_setting.start_color, RGBColor('off'))
        self.assertEqual(color_setting.dest_time,
                         color_setting.start_time + 2)
        self.assertEqual(color_setting.dest_color, RGBColor('red'))
        self.assertEqual(led1.get_color(), RGBColor('off'))
        self.assertFalse(color_setting.key)

        # advance to half way through the fade
        self.advance_time_and_run(1)

        self.assertTrue(led1.fade_in_progress)
        self.assertEqual(color_setting.priority, 0)
        self.assertEqual(color_setting.start_color, RGBColor('off'))
        self.assertEqual(color_setting.dest_time,
                         color_setting.start_time + 2)
        self.assertEqual(color_setting.dest_color, RGBColor('red'))
        self.assertEqual(led1.get_color(), RGBColor((127, 0, 0)))
        self.assertFalse(color_setting.key)
        self.assertLightColor("led1", [127, 0, 0])

        # advance to after the fade is done
        self.advance_time_and_run(2)

        self.assertFalse(led1.fade_in_progress)
        self.assertEqual(color_setting.priority, 0)
        self.assertEqual(color_setting.start_color, RGBColor('off'))
        self.assertEqual(color_setting.dest_color, RGBColor('red'))
        self.assertEqual(led1.get_color(), RGBColor('red'))
        self.assertFalse(color_setting.key)
        self.assertLightColor("led1", "red")

        led = self.machine.lights["led4"]
        self.assertEqual(1000, led.default_fade_ms)
        led.color('white')
        self.advance_time_and_run(.02)
        self.advance_time_and_run(.5)
        self.assertLightColor("led4", [132, 132, 132])
        self.advance_time_and_run(.5)
        self.assertLightColor("led4", [255, 255, 255])

    def test_restore_to_fade_in_progress(self):
        led1 = self.machine.lights["led1"]

        led1.color('red', fade_ms=4000, priority=50)
        self.advance_time_and_run(0.02)
        self.advance_time_and_run(1)

        # fade is 25% complete
        self.assertLightColor("led1", [65, 0, 0])

        # higher priority color which goes on top of fade
        led1.color('blue', key='test', priority=100)
        self.advance_time_and_run(1)
        self.assertLightColor("led1", 'blue')
        self.assertFalse(led1.fade_in_progress)

        led1.remove_from_stack_by_key('test')
        # should go back to the fade in progress, which is now 75% complete
        self.advance_time_and_run(1)
        self.assertLightColor("led1", [192, 0, 0])
        self.assertTrue(led1.fade_in_progress)

        # go to 1 sec after fade and make sure it finished
        self.advance_time_and_run(2)
        self.assertLightColor("led1", 'red')
        self.assertFalse(led1.fade_in_progress)

    def test_multiple_concurrent_fades(self):
        # start one fade, and while that's in progress, start a second fade.
        # the second fade should start from the wherever the current fade was.

        led1 = self.machine.lights["led1"]

        led1.color('red', fade_ms=4000, priority=50)
        self.advance_time_and_run(0.02)
        self.advance_time_and_run(1)

        # fade is 25% complete
        self.assertLightColor("led1", [65, 0, 0])

        # start a blue 2s fade
        led1.color('blue', key='test', fade_ms=2000, priority=100)

        # advance 1s, since we're half way to the blue fade from the 25% red,
        # we should now be at 12.5% red and 50% blue
        # Note: technically the red fade should continue even as it's being
        # faded to blue, but meh, we'll handle that with alpha channels in the
        # future
        self.advance_time_and_run(1)

        self.assertLightColor("led1", [33, 0, 127])

        # advance past the end
        self.advance_time_and_run(2)
        self.assertLightColor("led1", 'blue')
        self.assertFalse(led1.fade_in_progress)

    def test_color_correction(self):
        led = self.machine.lights["led_corrected"]
        led.color(RGBColor("white"))
        self.advance_time_and_run()
        # color is uncorrected
        self.assertLightColor("led_corrected", RGBColor("white"))
        # corrected color
        self.assertEqual(RGBColor([210, 184, 159]), led.color_correct(led.get_color()))
        # check hardware
        self.assertEqual(210 / 255.0, led.hw_drivers["red"][0].current_brightness)
        self.assertEqual(184 / 255.0, led.hw_drivers["green"][0].current_brightness)
        self.assertEqual(159 / 255.0, led.hw_drivers["blue"][0].current_brightness)

        led.color(RGBColor([128, 128, 128]))
        self.advance_time_and_run()
        self.assertLightColor("led_corrected", [128, 128, 128])
        self.assertEqual(RGBColor([96, 83, 70]), led.color_correct(led.get_color()))
        self.assertEqual(96 / 255.0, led.hw_drivers["red"][0].current_brightness)
        self.assertEqual(83 / 255.0, led.hw_drivers["green"][0].current_brightness)
        self.assertEqual(70 / 255.0, led.hw_drivers["blue"][0].current_brightness)

        led.color(RGBColor("black"))
        self.advance_time_and_run()
        self.assertLightColor("led_corrected", [0, 0, 0])
        self.assertEqual(RGBColor([0, 0, 0]), led.color_correct(led.get_color()))
        self.assertEqual(0 / 255.0, led.hw_drivers["red"][0].current_brightness)
        self.assertEqual(0 / 255.0, led.hw_drivers["green"][0].current_brightness)
        self.assertEqual(0 / 255.0, led.hw_drivers["blue"][0].current_brightness)

    def test_non_rgb_leds(self):
        # test bgr
        led = self.machine.lights["led2"]

        led.color(RGBColor((11, 23, 42)))
        self.advance_time_and_run(1)
        self.assertEqual(42 / 255, led.hw_drivers["blue"][0].current_brightness)
        self.assertEqual('led-2', led.hw_drivers["blue"][0].number)
        self.assertEqual(23 / 255, led.hw_drivers["green"][0].current_brightness)
        self.assertEqual('led-3', led.hw_drivers["green"][0].number)
        self.assertEqual(11 / 255, led.hw_drivers["red"][0].current_brightness)
        self.assertEqual('led-4', led.hw_drivers["red"][0].number)

        led = self.machine.lights["led_bgr_2"]
        led.color(RGBColor((11, 23, 42)))
        self.advance_time_and_run(1)
        self.assertEqual(42 / 255, led.hw_drivers["blue"][0].current_brightness)
        self.assertEqual('led-42-r', led.hw_drivers["blue"][0].number)
        self.assertEqual(23 / 255, led.hw_drivers["green"][0].current_brightness)
        self.assertEqual('led-42-g', led.hw_drivers["green"][0].number)
        self.assertEqual(11 / 255, led.hw_drivers["red"][0].current_brightness)
        self.assertEqual('led-42-b', led.hw_drivers["red"][0].number)

        # test rgbw
        led = self.machine.lights["led3"]

        led.color(RGBColor((11, 23, 42)))
        self.advance_time_and_run(1)
        self.assertLightColor("led2", [11, 23, 42])
        self.assertEqual(11 / 255, led.hw_drivers["white"][0].current_brightness)
        self.assertEqual('led-10', led.hw_drivers["white"][0].number)

        # test www light
        led = self.machine.lights["led_www"]
        led.on(128)
        self.advance_time_and_run(1)
        self.assertLightColor("led_www", [128, 128, 128])
        self.assertEqual(128 / 255, led.hw_drivers["white"][0].current_brightness)
        self.assertEqual('led-23-r', led.hw_drivers["white"][0].number)
        self.assertEqual(128 / 255, led.hw_drivers["white"][1].current_brightness)
        self.assertEqual('led-23-g', led.hw_drivers["white"][1].number)
        self.assertEqual(128 / 255, led.hw_drivers["white"][2].current_brightness)
        self.assertEqual('led-23-b', led.hw_drivers["white"][2].number)

    def test_brightness_correction(self):
        led = self.machine.lights["led1"]

        led.color(RGBColor((100, 100, 100)))
        self.advance_time_and_run(1)
        self.assertLightColor("led1", [100, 100, 100])
        self.assertEqual(100 / 255.0, led.hw_drivers["red"][0].current_brightness)
        self.assertEqual(100 / 255.0, led.hw_drivers["green"][0].current_brightness)
        self.assertEqual(100 / 255.0, led.hw_drivers["blue"][0].current_brightness)

        self.machine.variables.set_machine_var("brightness", 0.8)
        self.advance_time_and_run(.1)
        led.color(RGBColor((200, 200, 200)))    # color needs to change once for this to take effect
        self.advance_time_and_run(1)

        self.assertLightColor("led1", [200, 200, 200])
        self.assertEqual(160 / 255.0, led.hw_drivers["red"][0].current_brightness)
        self.assertEqual(160 / 255.0, led.hw_drivers["green"][0].current_brightness)
        self.assertEqual(160 / 255.0, led.hw_drivers["blue"][0].current_brightness)


class TestLightOnDriver(MpfTestCase):

    def get_config_file(self):
        return 'lights_on_drivers.yaml'

    def get_machine_path(self):
        return 'tests/machine_files/light/'

    def get_platform(self):
        # no force platform. we are testing the drivers platform
        return False

    def test_driver_platform(self):
        driver = self.machine.coils["coil_01"].hw_driver
        self.assertEqual("disabled", driver.state)
        self.machine.lights["light_on_driver"].on()
        self.assertEqual("enabled", driver.state)
        self.machine.lights["light_on_driver"].off()
        self.assertEqual("disabled", driver.state)



from unittest.mock import patch

from mpf.tests.MpfFakeGameTestCase import MpfFakeGameTestCase
from mpf.tests.MpfTestCase import test_config


class TestMatchMode(MpfFakeGameTestCase):

    def get_config_file(self):
        return 'config.yaml'

    def get_machine_path(self):
        return 'tests/machine_files/match_mode/'

    def get_platform(self):
        return 'smart_virtual'

    @test_config("config_highscore.yaml")
    def test_service_during_high_score(self):
        """Regression test for crash when service started during high score and cancels match."""
        self.post_event("add_credit")
        self.start_game()

        self.machine.game.player_list[0].score = 8000000
        self.drain_one_ball()
        self.advance_time_and_run(10)
        self.drain_one_ball()
        self.advance_time_and_run(10)
        self.drain_one_ball()
        self.advance_time_and_run(1)
        self.post_event("sw_service_enter_active")
        self.advance_time_and_run(1)

    def test_no_match(self):
        self.post_event("add_credit")
        self.assertMachineVarEqual(1, "credits_whole_num")

        self.start_game()
        self.assertMachineVarEqual(0, "credits_whole_num")

        self.machine.game.player.score += 1337

        self.mock_event("match_no_match")
        self.mock_event("match_has_match")

        with patch("mpf.modes.match.code.match.random.randint") as randint:
            randint.return_value = 50
            self.drain_all_balls()
            self.advance_time_and_run()
        self.assertGameIsNotRunning()
        self.assertEventCalled("match_no_match")
        self.assertEventNotCalled("match_has_match")

        self.assertMachineVarEqual(0, "credits_whole_num")

    def test_match(self):
        self.post_event("add_credit")
        self.assertMachineVarEqual(1, "credits_whole_num")

        self.start_game()
        self.assertMachineVarEqual(0, "credits_whole_num")

        self.machine.game.player.score += 1337

        self.mock_event("match_no_match")
        self.mock_event("match_has_match")

        with patch("mpf.modes.match.code.match.random.randint") as randint:
            randint.return_value = 5
            self.drain_all_balls()
            self.advance_time_and_run()
        self.assertGameIsNotRunning()
        self.assertEventNotCalled("match_no_match")
        self.assertEventCalled("match_has_match")

        self.assertMachineVarEqual(1, "credits_whole_num")



from unittest.mock import MagicMock

from mpf.tests.MpfTestCase import MpfTestCase


class TestShots(MpfTestCase):

    def get_config_file(self):
        return 'test_shots.yaml'

    def get_machine_path(self):
        return 'tests/machine_files/shots/'

    def setUp(self):
        super(TestShots, self).setUp()
        self.min_frame_time = 0.5

    def start_game(self):
        # shots only work in games so we have to do this a lot
        self.machine.playfield.add_ball = MagicMock()
        self.machine.events.post('game_start')
        self.advance_time_and_run(1.02)
        self.machine.game.balls_in_play = 1
        self.assertIsNotNone(self.machine.game)

    def stop_game(self):
        # stop game
        self.machine.game.end_game()
        self.advance_time_and_run()
        self.assertIsNone(self.machine.game)

    def test_block(self):
        self.mock_event("playfield_active")
        self.hit_and_release_switch("switch_3")
        self.advance_time_and_run(.1)
        self.assertEventCalled("playfield_active")

        self.start_game()
        self.assertEqual("unlit", self.machine.shots["shot_3"].state_name)

        self.hit_and_release_switch("switch_3")
        self.advance_time_and_run(.1)
        self.assertTrue(self.machine.shots["shot_3"].enabled)
        self.assertEqual("lit", self.machine.shots["shot_3"].state_name)

        self.machine.shots["shot_3"].reset()
        self.assertEqual("unlit", self.machine.shots["shot_3"].state_name)

        # Start the mode and make sure those shots load
        self.start_mode("mode1")

        self.assertTrue(self.machine.shots["shot_3"].enabled)
        self.assertTrue(self.machine.shots["mode1_shot_3"].enabled)
        self.assertEqual("unlit", self.machine.shots["shot_3"].state_name)
        self.assertEqual("mode1_one", self.machine.shots["mode1_shot_3"].state_name)

        self.hit_and_release_switch("switch_3")
        self.advance_time_and_run(.1)

        self.assertEqual("unlit", self.machine.shots["shot_3"].state_name)
        self.assertEqual("mode1_two", self.machine.shots["mode1_shot_3"].state_name)

    def test_loading_shots(self):
        # Make sure machine-wide shots load & mode-specific shots do not
        self.assertIn('shot_1', self.machine.shots)
        self.assertIn('shot_2', self.machine.shots)
        self.assertIn('shot_3', self.machine.shots)
        self.assertIn('shot_4', self.machine.shots)
        self.assertIn('led_1', self.machine.shots)

        self.assertFalse(self.machine.shots["mode1_shot_1"].enabled)

        self.start_game()

        # Start the mode and make sure those shots load
        self.start_mode("mode1")
        self.assertTrue(self.machine.shots["mode1_shot_1"].enabled)

        self.assertIn('shot_1', self.machine.shots)
        self.assertIn('shot_2', self.machine.shots)
        self.assertIn('shot_3', self.machine.shots)
        self.assertIn('shot_4', self.machine.shots)
        self.assertIn('led_1', self.machine.shots)

        # Stop the mode and make sure those shots go away
        self.stop_mode("mode1")
        self.assertFalse(self.machine.shots["mode1_shot_1"].enabled)

        self.assertIn('shot_1', self.machine.shots)
        self.assertIn('shot_2', self.machine.shots)
        self.assertIn('shot_3', self.machine.shots)
        self.assertIn('shot_4', self.machine.shots)
        self.assertIn('led_1', self.machine.shots)

    def test_mode_priorities(self):
        self.start_game()

        # Start the mode
        self.start_mode("mode1")
        # check shot states
        self.assertTrue(self.machine.shots["mode1_shot_2"].enabled)
        self.assertFalse(self.machine.shots["mode2_shot_2"].enabled)
        self.assertLightColor("light_2", "aliceblue")
        self.hit_and_release_switch('switch_2')
        self.assertLightColor("light_2", "antiquewhite")

        self.start_mode("mode2")
        self.assertTrue(self.machine.shots["mode1_shot_2"].enabled)
        self.assertTrue(self.machine.shots["mode2_shot_2"].enabled)

        # mode2 takes priority
        self.assertLightColor("light_2", "red")

        # Stop the mode
        self.stop_mode("mode2")
        self.assertTrue(self.machine.shots["mode1_shot_2"].enabled)
        self.assertFalse(self.machine.shots["mode2_shot_2"].enabled)

        # check if color returns to mode1_shot_2 color
        self.assertLightColor("light_2", "antiquewhite")

    def test_hits(self):
        self.assertFalse(self.machine.shots["mode1_shot_1"].enabled)

        self.shot_1_hit = MagicMock()
        self.shot_1_default_hit = MagicMock()
        self.shot_1_default_unlit_hit = MagicMock()
        self.mode1_shot_1_hit = MagicMock()
        self.machine.events.add_handler('shot_1_hit', self.shot_1_hit)
        self.machine.events.add_handler('shot_1_default_hit',
                                        self.shot_1_default_hit)
        self.machine.events.add_handler('shot_1_default_unlit_hit',
                                        self.shot_1_default_unlit_hit)
        self.machine.events.add_handler('mode1_shot_1_hit',
                                        self.mode1_shot_1_hit)

        # make sure shot does not work with no game in progress
        self.hit_and_release_switch('switch_1')
        self.advance_time_and_run()
        self.shot_1_hit.assert_not_called()

        self.start_game()
        self.start_mode("mode1")
        self.assertTrue(self.machine.shots["mode1_shot_1"].enabled)

        # hit shot_1, test all three event variations
        self.hit_and_release_switch('switch_1')
        self.advance_time_and_run()

        self.shot_1_hit.assert_called_once_with(profile='default',
                                                state='unlit', advancing=True)
        self.shot_1_default_hit.assert_called_once_with(profile='default',
                                                        state='unlit', advancing=True)
        self.shot_1_default_unlit_hit.assert_called_once_with(
            profile='default', state='unlit', advancing=True)

        # hit the mode shot and make sure it was called
        self.hit_and_release_switch('switch_3')
        self.advance_time_and_run()
        self.mode1_shot_1_hit.assert_called_once_with(profile='default',
                                                      state='unlit', advancing=True)

        # stop the mode
        self.machine.modes["mode1"].stop()
        self.advance_time_and_run()

        # hit the mode shot and make sure it was not called
        self.mode1_shot_1_hit = MagicMock()
        self.hit_and_release_switch('switch_3')
        self.advance_time_and_run()
        self.mode1_shot_1_hit.assert_not_called()

        # stop the game (should not crash)
        self.stop_game()

        # hit the shot and make sure it was not called again
        self.hit_and_release_switch('switch_1')
        self.advance_time_and_run()
        self.mode1_shot_1_hit.assert_not_called()

    def test_shot_with_multiple_switches(self):
        self.shot_15_hit = MagicMock()
        self.machine.events.add_handler('shot_15_hit', self.shot_15_hit)
        self.start_game()

        # hit shot_15 via switch_13
        self.hit_and_release_switch('switch_13')
        self.shot_15_hit.assert_called_once_with(profile='default',
                                                 state='unlit', advancing=True)

        # hit shot_15 via switch_14
        self.shot_15_hit.reset_mock()
        self.hit_and_release_switch('switch_14')
        self.shot_15_hit.assert_called_once_with(profile='default',
                                                 state='lit', advancing=False)

    def test_shot_with_delay(self):
        self.mock_event("shot_delay_hit")
        self.mock_event("shot_delay_same_switch_hit")
        self.start_game()

        # test delay at the beginning. should not count
        self.hit_and_release_switch("s_delay")
        self.advance_time_and_run(.5)
        self.hit_and_release_switch("switch_1")
        self.advance_time_and_run(1)
        self.assertEventNotCalled("shot_delay_hit")
        self.advance_time_and_run(3)

        # test shot with same switch for delay and hit
        self.hit_and_release_switch("switch_15")
        self.advance_time_and_run(.5)
        self.assertEventCalled("shot_delay_same_switch_hit")
        self.mock_event("shot_delay_same_switch_hit")
        self.hit_and_release_switch("switch_15")
        self.advance_time_and_run(.5)
        self.assertEventNotCalled("shot_delay_same_switch_hit")

        self.advance_time_and_run(3)
        self.hit_and_release_switch("switch_15")
        self.advance_time_and_run(.5)
        self.assertEventCalled("shot_delay_same_switch_hit")
        self.mock_event("shot_delay_same_switch_hit")

        # test that shot works without delay
        self.hit_and_release_switch("switch_1")
        self.advance_time_and_run(.5)
        self.assertEventCalled("shot_delay_hit")
        self.mock_event("shot_delay_hit")

        self.hit_and_release_switch("switch_1")
        self.advance_time_and_run(.5)
        self.assertEventCalled("shot_delay_hit")
        self.mock_event("shot_delay_hit")
        self.hit_and_release_switch("s_delay")

        self.machine.modes["base2"].stop()
        self.advance_time_and_run()
        self.hit_and_release_switch("switch_1")
        self.advance_time_and_run(.5)
        self.assertEventNotCalled("shot_delay_hit")

        self.machine.modes["base2"].start()
        self.advance_time_and_run()
        self.hit_and_release_switch("switch_1")
        self.advance_time_and_run(.5)
        self.assertEventCalled("shot_delay_hit")

        self.advance_time_and_run(3)
        self.hit_and_release_switch("switch_15")
        self.advance_time_and_run(.5)
        self.assertEventCalled("shot_delay_same_switch_hit")
        self.mock_event("shot_delay_same_switch_hit")

    def test_profile_advancing_no_loop(self):
        self.start_game()
        self.mock_event("shot_27_hit")
        # unlit and two states in the beginning
        self.assertEqual(2, len(self.machine.shots["shot_1"].config['profile'].config['states']))
        self.assertEqual("unlit", self.machine.shots["shot_1"].state_name)
        self.assertPlaceholderEvaluates("unlit", "device.shots.shot_1.state_name")
        self.assertPlaceholderEvaluates(0, "device.shots.shot_1.state")

        # one hit and it lits
        self.hit_and_release_switch("switch_1")
        self.assertEqual("lit", self.machine.shots["shot_1"].state_name)
        self.assertEqual(1, self.machine.game.player["shot_shot_1"])
        self.assertPlaceholderEvaluates("lit", "device.shots.shot_1.state_name")
        self.assertPlaceholderEvaluates(1, "device.shots.shot_1.state")

        # it stays lit
        self.hit_and_release_switch("switch_1")
        self.assertEqual("lit", self.machine.shots["shot_1"].state_name)
        self.assertEqual(1, self.machine.game.player["shot_shot_1"])
        self.assertEventCalled("shot_27_hit")

    def test_shots_with_events(self):
        self.start_game()
        self.assertModeRunning("base2")
        self.mock_event("shot_28_hit")
        self.post_event("event1")
        self.assertEventCalled("shot_28_hit")

    def test_profile_advancing_with_loop(self):
        self.start_game()

        self.assertEqual(3, len(self.machine.shots["shot_2"].config['profile'].config['states']))

        self.assertEqual("one", self.machine.shots["shot_2"].state_name)

        self.hit_and_release_switch("switch_2")
        self.assertEqual("two", self.machine.shots["shot_2"].state_name)
        self.assertEqual(1, self.machine.game.player["shot_shot_2"])

        self.hit_and_release_switch("switch_2")
        self.assertEqual("three", self.machine.shots["shot_2"].state_name)
        self.assertEqual(2, self.machine.game.player["shot_shot_2"])

        self.hit_and_release_switch("switch_2")
        self.assertEqual("one", self.machine.shots["shot_2"].state_name)
        self.assertEqual(0, self.machine.game.player["shot_shot_2"])

        self.hit_and_release_switch("switch_2")
        self.assertEqual("two", self.machine.shots["shot_2"].state_name)
        self.assertEqual(1, self.machine.game.player["shot_shot_2"])

    def test_default_show_light(self):
        self.start_game()
        self.assertLightChannel("light_4", 0)

        self.hit_and_release_switch("switch_5")
        self.advance_time_and_run()
        self.assertLightChannel("light_4", 255)

    def test_default_show_lights(self):
        self.start_game()
        self.assertLightChannel("light_5", 0)
        self.assertLightChannel("light_6", 0)

        self.hit_and_release_switch("switch_6")
        self.advance_time_and_run()
        self.assertLightChannel("light_5", 255)
        self.assertLightChannel("light_6", 255)

    def test_default_show_led(self):
        self.start_game()
        self.assertLightColor("led_4", "off")

        self.hit_and_release_switch("switch_7")
        self.advance_time_and_run()

        self.assertLightColor("led_4", "white")

    def test_default_show_leds(self):
        self.start_game()
        self.assertLightColor("led_5", "off")
        self.assertLightColor("led_6", "off")

        self.hit_and_release_switch("switch_8")
        self.advance_time_and_run()

        self.assertLightColor("led_5", "white")
        self.assertLightColor("led_6", "white")

    def test_show_in_shot_profile_root(self):
        self.start_game()
        self.assertLightColor("led_3", "red")

        # make sure the show is not auto advancing
        self.advance_time_and_run(5)
        self.assertLightColor("led_3", "red")

        self.advance_time_and_run(5)
        self.assertLightColor("led_3", "red")

        self.hit_and_release_switch("switch_9")
        self.advance_time_and_run()
        self.assertLightColor("led_3", "orange")

        self.hit_and_release_switch("switch_9")
        self.advance_time_and_run()
        self.assertLightColor("led_3", "yellow")

        self.hit_and_release_switch("switch_9")
        self.advance_time_and_run()
        self.assertLightColor("led_3", "green")

        # make sure it stays on green
        self.advance_time_and_run(5)
        self.assertLightColor("led_3", "green")

    def test_show_in_step(self):
        self.start_game()
        # start_game() advances the time 1 sec, so by now we're already on
        # step 2 of the rainbow show
        self.assertLightColor("led_11", "orange")

        # make sure show is advancing on its own
        self.advance_time_and_run(1)
        self.assertLightColor("led_11", "yellow")

        # hit the shot, changes to show1
        self.hit_and_release_switch("switch_11")
        self.advance_time_and_run(0.1)
        self.assertLightColor("led_11", "aliceblue")

        # make sure show is advancing on its own
        self.advance_time_and_run(1)
        self.assertLightColor("led_11", "antiquewhite")

    def test_combined_show_in_profile_root_and_step(self):
        # tests a show defined in a profile root which is used for most steps,
        # but a separate show in certain steps that is used just for that step

        self.start_game()
        self.advance_time_and_run()

        # we're on step 1
        self.assertLightColor("led_12", "red")

        # step 2
        self.hit_and_release_switch("switch_12")
        self.advance_time_and_run()
        self.assertLightColor("led_12", "orange")

        # since this is a root show, it should not be advancing on its own, so
        # advance the time a few times and make sure the led doesn't change
        self.advance_time_and_run()
        self.assertLightColor("led_12", "orange")

        self.advance_time_and_run()
        self.assertLightColor("led_12", "orange")

        self.advance_time_and_run()
        self.assertLightColor("led_12", "orange")

        # step 3 is rainbow 2 show, so make sure it switches
        self.hit_and_release_switch("switch_12")
        self.advance_time_and_run(0.02)
        self.assertLightColor("led_12", "aliceblue")

        # since this is a show in a step, it should be auto advancing, so keep
        # checking every sec to make sure the colors are changing
        self.advance_time_and_run(1)
        self.assertLightColor("led_12", "antiquewhite")

        self.advance_time_and_run(1)
        self.assertLightColor("led_12", "aquamarine")

        self.advance_time_and_run(1)
        self.assertLightColor("led_12", "azure")

        # it should loop
        self.advance_time_and_run(1)
        self.assertLightColor("led_12", "aliceblue")

        self.advance_time_and_run(1)
        self.assertLightColor("led_12", "antiquewhite")

        # hit the switch, should advance to step 4, which is back to the
        # rainbow show
        self.hit_and_release_switch("switch_12")
        self.advance_time_and_run(1)
        self.assertLightColor("led_12", "green")

        # show should not be advancing without a hit
        self.advance_time_and_run(1)
        self.assertLightColor("led_12", "green")

        # hit to verify advance
        self.hit_and_release_switch("switch_12")
        self.advance_time_and_run(1)
        self.assertLightColor("led_12", "blue")

    def test_show_ending_no_loop(self):
        # tests that if a show is set to loops: 0, that it truly stops on the
        # last step. Note that loops here is really a setting of the show
        # player, so if this works, that means that all of the show player
        # settings will work in a show tied to a shot.

        self.start_game()
        self.assertLightColor("led_14", "orange")

        self.advance_time_and_run(1)
        self.assertLightColor("led_14", "yellow")

        self.advance_time_and_run(1)
        self.assertLightColor("led_14", "green")

        self.advance_time_and_run(1)
        self.assertLightColor("led_14", "blue")

        self.advance_time_and_run(1)
        self.assertLightColor("led_14", "purple")

        # show is done, but hold is true by default in shows started from shots
        # so make sure the led stays in its final state
        self.advance_time_and_run(1)
        self.assertLightColor("led_14", "purple")

    def test_enable_persist(self):
        self.start_game()

        self.start_mode("mode1")
        self.assertFalse(self.machine.shots["mode1_shot_17"].enabled)
        self.assertTrue(self.machine.shots["mode1_shot_1"].enabled)
        self.assertPlaceholderEvaluates(True, "device.shots.mode1_shot_1.enabled")

        self.stop_mode("mode1")
        self.start_mode("mode1")
        self.assertFalse(self.machine.shots["mode1_shot_17"].enabled)
        self.assertTrue(self.machine.shots["mode1_shot_1"].enabled)

        self.post_event("custom_enable_17")
        self.assertTrue(self.machine.shots["mode1_shot_17"].enabled)
        self.post_event("custom_disable_1")
        self.assertFalse(self.machine.shots["mode1_shot_1"].enabled)
        self.assertPlaceholderEvaluates(False, "device.shots.mode1_shot_1.enabled")

        self.stop_mode("mode1")
        self.assertFalse(self.machine.shots["mode1_shot_17"].enabled)
        self.assertFalse(self.machine.shots["mode1_shot_1"].enabled)

        self.start_mode("mode1")
        self.assertTrue(self.machine.shots["mode1_shot_17"].enabled)
        self.assertFalse(self.machine.shots["mode1_shot_1"].enabled)

    def test_control_events(self):
        # test control events from machine-wide shot
        shot16 = self.machine.shots["shot_16"]
        self.mock_event("shot_16_hit")
        self.start_game()

        # Since this shot has custom enable events, it should not be enabled on
        # game start
        self.assertFalse(shot16.enabled)

        # Also verify hit event does not register a hit
        self.machine.events.post('custom_hit_16')
        self.advance_time_and_run()
        self.assertEqual(0, self._events["shot_16_hit"])

        # test enabling via event
        self.machine.events.post('custom_enable_16')
        self.advance_time_and_run()
        self.assertTrue(shot16.enabled)

        # test hit event
        self.machine.events.post('custom_hit_16')
        self.advance_time_and_run()
        self.assertEqual(1, self._events["shot_16_hit"])
        self.assertEqual('lit', shot16.state_name)

        # test reset event
        self.machine.events.post('custom_reset_16')
        self.advance_time_and_run()
        self.assertEqual(shot16.state_name, 'unlit')

        # test advance event
        self.machine.events.post('custom_advance_16')
        self.advance_time_and_run()
        # hit should not be posted again (still 1 from before)
        self.assertEqual(1, self._events["shot_16_hit"])
        # profile should have advanced though
        self.assertEqual(shot16.state_name, 'lit')

        # test disable event
        self.machine.events.post('custom_disable_16')
        self.advance_time_and_run()
        self.assertFalse(shot16.enabled)

        # hit the shot to make sure it's really disabled
        self.machine.events.post('custom_hit_16')
        self.advance_time_and_run()
        self.assertEqual(1, self._events["shot_16_hit"])  # still 1 from before
        self.assertEqual(shot16.state_name, 'lit')

        self.post_event("custom_enable_16")
        self.assertTrue(self.machine.shots["shot_16"].enabled)

        self.post_event('custom_hit_16')
        self.assertEqual(2, self._events["shot_16_hit"])
        self.assertEqual(shot16.state_name, 'lit')

        # test restart event
        self.machine.events.post('custom_disable_16')
        self.advance_time_and_run()
        # make sure we are disabled and advanced in the profile
        self.assertEqual(shot16.state_name, 'lit')
        self.assertFalse(self.machine.shots["shot_16"].enabled)
        self.machine.events.post('custom_restart_16')
        self.advance_time_and_run()
        self.assertEqual(shot16.state_name, 'unlit')
        self.assertTrue(self.machine.shots["shot_16"].enabled)

        # mode1 is not active, so make sure none of the events from
        # mode1_shot_17

        self.assertFalse(self.machine.shots["mode1_shot_17"].enabled)
        self.mock_event("mode1_shot_17_hit")

        # Also verify hit event does not register a hit
        self.machine.events.post('custom_hit_17')
        self.advance_time_and_run()
        self.assertEqual(0, self._events["mode1_shot_17_hit"])

        # test enabling via event, shot should not work
        self.machine.events.post('custom_enable_17')
        self.advance_time_and_run()

        self.machine.events.post('custom_hit_17')
        self.advance_time_and_run()
        self.assertEqual(0, self._events["mode1_shot_17_hit"])

        # test reset event, nothing should happen
        self.machine.events.post('custom_reset_17')
        self.advance_time_and_run()

        # test disable event, nothing should happen
        self.machine.events.post('custom_disable_17')
        self.advance_time_and_run()

        # start the mode
        self.machine.modes["mode1"].start()
        self.advance_time_and_run()

        shot17 = self.machine.shots["mode1_shot_17"]

        # Since this shot has custom enable events, it should not be enabled on
        # game start
        self.assertFalse(shot17.enabled)

        # Also verify hit event does not register a hit
        self.machine.events.post('custom_hit_17')
        self.advance_time_and_run()
        self.assertEqual(0, self._events["mode1_shot_17_hit"])

        # test enabling via event
        self.machine.events.post('custom_enable_17')
        self.advance_time_and_run()
        self.assertTrue(shot17.enabled)

        # test hit event
        self.machine.events.post('custom_hit_17')
        self.advance_time_and_run()
        self.assertEqual(1, self._events["mode1_shot_17_hit"])
        self.assertEqual(shot17.state_name, 'lit')

        # test reset event
        self.machine.events.post('custom_reset_17')
        self.advance_time_and_run()
        self.assertEqual(shot17.state_name, 'unlit')

        # test disable event
        self.machine.events.post('custom_disable_17')
        self.advance_time_and_run()
        self.assertFalse(shot17.enabled)

        # hit the shot to make sure it's really disabled
        self.machine.events.post('custom_hit_17')
        self.advance_time_and_run()
        # since it's disabled, there should still only be 1 from before
        self.assertEqual(1, self._events["mode1_shot_17_hit"])
        self.assertEqual(shot17.state_name, 'unlit')

    def test_advance(self):
        self.mock_event("shot_17_hit")
        shot17 = self.machine.shots["shot_17"]

        self.start_game()

        # shot profile config has advance_on_hit: false
        self.hit_and_release_switch("switch_17")
        # verify it was hit, but it didn't advance
        self.assertEqual(1, self._events["shot_17_hit"])
        self.assertEqual("one", shot17.state_name)

        # again
        self.hit_and_release_switch("switch_17")
        self.assertEqual(2, self._events["shot_17_hit"])
        self.assertEqual("one", shot17.state_name)

        # manual advance
        shot17.advance()
        shot17.advance()
        self.assertEqual("three", shot17.state_name)

        # hit still doesn't advance
        self.hit_and_release_switch("switch_17")
        self.assertEqual(3, self._events["shot_17_hit"])
        self.assertEqual("three", shot17.state_name)

        # disable the shot, advance should be disabled too
        shot17.disable()
        shot17.advance()
        self.assertEqual(2, self.machine.game.player.shot_shot_17)

        # though we can force it to advance
        shot17.advance(force=True)
        self.assertEqual(3, self.machine.game.player.shot_shot_17)

    def test_show_when_disabled(self):
        # first test show_when_disabled == true

        shot19 = self.machine.shots["shot_19"]

        self.start_game()
        # shot19 config has enable_events: none, so it should be disabled
        self.assertFalse(shot19.enabled)

        # start_game() includes a 1 sec advance time, so by now this show is
        # already on step 2
        self.assertLightColor("led_19", 'orange')

        # make sure the show keeps running
        self.advance_time_and_run()
        self.assertLightColor("led_19", 'yellow')

        # enable the shot
        shot19.enable()
        self.advance_time_and_run(.1)

        # show should still be at the same step
        self.assertLightColor("led_19", 'yellow')

        # but it should also still be running
        self.advance_time_and_run(1)
        self.assertLightColor("led_19", 'green')

        # hit the shot
        shot19.hit()
        self.advance_time_and_run(.1)

        # should switch to the second show
        self.assertLightColor("led_19", 'aliceblue')

        # and that show should be running
        self.advance_time_and_run()
        self.assertLightColor("led_19", 'antiquewhite')

        # disable the shot
        shot19.disable()
        self.advance_time_and_run(.1)

        # color should not change
        self.assertLightColor("led_19", 'antiquewhite')

        # and show should still be running
        self.advance_time_and_run(1)
        self.assertLightColor("led_19", 'aquamarine')

    def test_no_show_when_disabled(self):
        shot20 = self.machine.shots["shot_20"]

        self.start_game()

        # shot20 should be disabled
        self.assertFalse(shot20.enabled)

        # make sure the show is not running and not affecting the LED
        self.assertLightColor("led_20", 'off')

        # enable the shot, show should start
        shot20.enable()

        self.assertFalse(shot20.config['profile'].config['show_when_disabled'])
        self.assertTrue(shot20.enabled)

        self.advance_time_and_run(.1)
        self.assertLightColor("led_20", 'red')

        # make sure show is advancing
        self.advance_time_and_run(1)
        self.assertLightColor("led_20", 'orange')

        # hit the shot, show should switch
        shot20.hit()
        self.advance_time_and_run(.1)

        self.assertLightColor("led_20", 'aliceblue')

        # and that show should be running
        self.advance_time_and_run()
        self.assertLightColor("led_20", 'antiquewhite')

        # disable the shot
        shot20.disable()
        self.advance_time_and_run()

        # LEDs should be off since show_when_disabled == false
        self.assertLightColor("led_20", 'off')

    def test_hold_true(self):
        self.start_game()
        self.assertLightColor("led_24", 'orange')

        # advance the time past the end of the show and make sure that the
        # led is still on
        # all these separate entries are needed due to the way the tests run
        self.advance_time_and_run(1)
        self.advance_time_and_run(1)
        self.advance_time_and_run(1)
        self.advance_time_and_run(1)
        self.advance_time_and_run(1)
        self.advance_time_and_run(1)
        self.advance_time_and_run(1)
        self.advance_time_and_run(1)
        self.advance_time_and_run(1)
        self.advance_time_and_run(1)
        self.advance_time_and_run(1)
        self.assertLightColor("led_24", 'purple')

    def test_hold_false(self):
        self.start_game()
        self.assertLightColor("led_25", 'orange')

        # advance the time past the end of the show and make sure that the
        # led is off
        self.advance_time_and_run(1)
        self.advance_time_and_run(1)
        self.advance_time_and_run(1)
        self.advance_time_and_run(1)
        self.advance_time_and_run(1)
        self.advance_time_and_run(1)
        self.advance_time_and_run(1)
        self.advance_time_and_run(1)
        self.advance_time_and_run(1)
        self.advance_time_and_run(1)
        self.advance_time_and_run(1)
        self.assertLightColor("led_25", 'off')

    def test_show_restore_in_mode(self):
        self.start_game()

        self.assertLightColor("led_27", "black")

        self.machine.modes["mode2"].start()
        self.advance_time_and_run()

        # step1 red
        self.assertLightColor("led_27", "red")

        self.hit_and_release_switch("switch_27")
        self.advance_time_and_run()

        # step2 orange
        self.assertLightColor("led_27", "orange")

        self.machine.modes["mode2"].stop()
        self.advance_time_and_run()

        # mode stopped. led off
        self.assertLightColor("led_27", "black")

        self.machine.modes["mode2"].start()
        self.advance_time_and_run()

        # back to step2. orange
        self.assertLightColor("led_27", "orange")

        self.hit_and_release_switch("switch_27")
        self.advance_time_and_run()

        # step3
        self.assertLightColor("led_27", "yellow")

    def test_show_restore_in_mode_start_step(self):
        # same as previous test but with a profile with start steps
        self.start_game()

        self.assertLightColor("led_28", "black")

        self.machine.modes["mode2"].start()
        self.advance_time_and_run()
        self.assertTrue(self.machine.shots["mode2_shot_rainbow_start_step"].enabled)

        # step1 red
        self.assertLightColor("led_28", "red")
        self.assertEqual("red", self.machine.shots["mode2_shot_rainbow_start_step"].state_name)

        self.hit_and_release_switch("switch_28")
        self.advance_time_and_run()

        # step2 orange
        self.assertLightColor("led_28", "orange")
        self.assertEqual("orange", self.machine.shots["mode2_shot_rainbow_start_step"].state_name)

        self.machine.modes["mode2"].stop()
        self.advance_time_and_run()

        # mode stopped. led off
        self.assertLightColor("led_28", "black")
        self.assertFalse(self.machine.shots["mode2_shot_rainbow_start_step"].enabled)

        self.machine.modes["mode2"].start()
        self.advance_time_and_run()

        # back to step2. orange
        self.assertEqual("orange", self.machine.shots["mode2_shot_rainbow_start_step"].state_name)
        self.assertLightColor("led_28", "orange")
        self.assertTrue(self.machine.shots["mode2_shot_rainbow_start_step"].enabled)

        self.hit_and_release_switch("switch_28")
        self.advance_time_and_run()

        # step3
        self.assertEqual("yellow", self.machine.shots["mode2_shot_rainbow_start_step"].state_name)
        self.assertLightColor("led_28", "yellow")

    def test_show_tokens_in_shot(self):
        """Test show tokens in shots and shot_profiles."""
        self.start_game()
        self.machine.modes["mode2"].start()

        self.assertLightColor("led_29", "black")

        # make sure nothing crashes when variables are missing
        self.post_event("mode2_shot_show_tokens_enable")
        self.assertLightColor("led_29", "black")
        self.assertEqual("one", self.machine.shots["mode2_shot_show_tokens"].state_name)

        self.post_event("mode2_shot_show_tokens_advance")
        self.assertLightColor("led_29", "black")
        self.assertEqual("two", self.machine.shots["mode2_shot_show_tokens"].state_name)

        # set variables
        self.machine.variables.set_machine_var("leds", "led_29")
        self.machine.variables.set_machine_var("color1", "red")
        self.machine.variables.set_machine_var("color2", "green")
        self.machine.variables.set_machine_var("color3", "blue")

        # try again
        self.post_event("mode2_shot_show_tokens_reset")

        self.assertLightColor("led_29", "red")
        self.assertEqual("one", self.machine.shots["mode2_shot_show_tokens"].state_name)

        self.post_event("mode2_shot_show_tokens_advance")
        self.assertLightColor("led_29", "green")
        self.assertEqual("two", self.machine.shots["mode2_shot_show_tokens"].state_name)

        self.post_event("mode2_shot_show_tokens_advance")
        self.assertLightColor("led_29", "blue")
        self.assertEqual("three", self.machine.shots["mode2_shot_show_tokens"].state_name)

    def test_jump(self):
        """Test jumping shots and shot_profiles."""
        self.start_game()
        self.machine.modes["mode2"].start()
        shot = self.machine.device_manager.collections["shots"]["mode2_shot_changing_profile"]

        # Initial color of the light for this profile
        self.assertLightColor("led_20", "yellow")

        # Change the profile and jump without force_show
        shot.config['profile'] = self.machine.device_manager.collections["shot_profiles"]['changing_profile_two']
        shot.jump(0)
        # State is the same, no color change
        self.assertLightColor("led_20", "yellow")

        # Jump and force_show
        shot.jump(0, True, True)
        # Should see the color of the new profile at the same state
        self.assertLightColor("led_20", "purple")
